import os
import re


def fuzzy_match(text: str, pattern: str) -> tuple[bool, float]:
    """
    Match text against a pattern using a fuzzy search algorithm similar to VSCode's Ctrl+P.

    The pattern is split into tokens by whitespace and path separators.
    Each token must be found in the text (in order).

    Args:
        text: The string to search in.
        pattern: The search pattern (e.g., "src main" or "util/io").

    Returns:
        A tuple (matched, score).
        - matched: True if the pattern matches the text.
        - score: A float representing the match quality (lower is better).
    """
    text_cmp = text.lower()
    # Normalize pattern -> tokens split on path separators or whitespace
    search_pattern = pattern.strip()
    tokens = (
        [t for t in re.split(rf"[{re.escape(os.path.sep)}\s]+", search_pattern) if t]
        if search_pattern
        else []
    )
    tokens = [t.lower() for t in tokens]
    if not tokens:
        return True, 0.0
    last_pos = 0
    score = 0.0
    for token in tokens:
        # try contiguous substring search first
        idx = text_cmp.find(token, last_pos)
        if idx != -1:
            # good match: reward contiguous early matches
            score += idx  # smaller idx preferred
            last_pos = idx + len(token)
        else:
            # fallback to subsequence matching
            res = _find_subsequence_range(text_cmp, token, last_pos)
            if res is None:
                return False, 0.0

            pos, end_pos = res

            # subsequence match is less preferred than contiguous substring
            score += pos + 0.5 * len(token)
            last_pos = end_pos
    # prefer shorter texts when score ties, so include length as tiebreaker
    score += 0.01 * len(text)
    return True, score


def _find_subsequence_range(
    hay: str, needle: str, start: int = 0
) -> tuple[int, int] | None:
    """
    Try to locate needle in hay as a subsequence starting at `start`.
    Returns (start_index, end_index) where end_index is the index AFTER the last matched character.
    """
    if not needle:
        return start, start
    i = start
    j = 0
    first_pos = None
    while i < len(hay) and j < len(needle):
        if hay[i] == needle[j]:
            if first_pos is None:
                first_pos = i
            j += 1
        i += 1

    if j == len(needle):
        return first_pos, i
    return None



from fastapi.testclient import TestClient
from my_app_name.main import app
from my_app_name.test._util.access_token import get_admin_access_token


def test_create_my_entity():
    client = TestClient(app, base_url="http://localhost")
    access_token = get_admin_access_token()
    new_my_entity_data = {
        "my_column": "new-my-entity",
    }
    response = client.post(
        "/api/v1/my-entities",
        json=new_my_entity_data,
        headers={"Authorization": f"Bearer {access_token}"},
    )
    assert response.status_code == 200
    response_data = response.json()
    assert response_data.get("id") is not None
    assert response_data.get("id") != ""
    assert response_data.get("my_column") == "new-my-entity"


def test_create_my_entity_bulk():
    client = TestClient(app, base_url="http://localhost")
    access_token = get_admin_access_token()
    new_first_my_entity_data = {
        "my_column": "new-my-entity-bulk-1",
    }
    new_second_my_entity_data = {
        "my_column": "new-my-entity-bulk-2",
    }
    new_my_entity_data = [new_first_my_entity_data, new_second_my_entity_data]
    response = client.post(
        "/api/v1/my-entities/bulk",
        json=new_my_entity_data,
        headers={"Authorization": f"Bearer {access_token}"},
    )
    assert response.status_code == 200
    response_data = response.json()
    assert len(response_data) == 2
    # Id should not be empty
    assert response_data[0] is not None
    assert response_data[0] != ""
    assert response_data[1] is not None
    assert response_data[1] != ""
    # Data should match
    assert new_first_my_entity_data["my_column"] in [
        row["my_column"] for row in response_data
    ]
    assert new_second_my_entity_data["my_column"] in [
        row["my_column"] for row in response_data
    ]



from zrb.builtin.group import llm_group
from zrb.config.config import CFG
from zrb.input.bool_input import BoolInput
from zrb.input.str_input import StrInput
from zrb.llm.app.confirmation.replace_confirmation import replace_confirmation
from zrb.llm.history_processor.summarizer import create_summarizer_history_processor
from zrb.llm.note.manager import NoteManager
from zrb.llm.prompt.claude_compatibility import (
    create_claude_compatibility_prompt,
)
from zrb.llm.prompt.compose import PromptManager, new_prompt
from zrb.llm.prompt.default import (
    get_mandate_prompt,
    get_persona_prompt,
)
from zrb.llm.prompt.note import create_note_prompt
from zrb.llm.prompt.system_context import system_context
from zrb.llm.prompt.zrb import create_zrb_prompt
from zrb.llm.skill.manager import SkillManager
from zrb.llm.task.llm_chat_task import LLMChatTask
from zrb.llm.tool.bash import run_shell_command
from zrb.llm.tool.code import analyze_code
from zrb.llm.tool.file import (
    analyze_file,
    glob_files,
    list_files,
    read_file,
    read_files,
    replace_in_file,
    search_files,
    write_file,
    write_files,
)
from zrb.llm.tool.mcp import load_mcp_config
from zrb.llm.tool.note import create_note_tools
from zrb.llm.tool.skill import create_activate_skill_tool
from zrb.llm.tool.web import open_web_page, search_internet
from zrb.llm.tool.zrb_task import create_list_zrb_task_tool, create_run_zrb_task_tool
from zrb.runner.cli import cli

skill_manager = SkillManager()
note_manager = NoteManager()

llm_chat = LLMChatTask(
    name="chat",
    description="🤖 Chat with your AI Assistant",
    input=[
        StrInput("message", "Message", allow_empty=True, always_prompt=False),
        StrInput(
            "session", "Conversation Session", allow_empty=True, always_prompt=False
        ),
        BoolInput(
            "yolo", "YOLO Mode", default=False, allow_empty=True, always_prompt=False
        ),
        StrInput("attach", "Attachments", allow_empty=True, always_prompt=False),
        BoolInput(
            "interactive",
            "Interactive Mode",
            default=True,
            allow_empty=True,
            always_prompt=False,
        ),
    ],
    yolo="{ctx.input.yolo}",
    message="{ctx.input.message}",
    conversation_name="{ctx.input.session}",
    interactive="{ctx.input.interactive}",
    history_processors=[
        create_summarizer_history_processor(
            token_threshold=CFG.LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD,
            summary_window=CFG.LLM_HISTORY_SUMMARIZATION_WINDOW,
        )
    ],
    prompt_manager=PromptManager(),
    ui_ascii_art=lambda ctx: CFG.LLM_ASSISTANT_ASCII_ART,
    ui_assistant_name=lambda ctx: CFG.LLM_ASSISTANT_NAME,
    ui_greeting=lambda ctx: f"{CFG.LLM_ASSISTANT_NAME}\n{CFG.LLM_ASSISTANT_JARGON}",
    ui_jargon=lambda ctx: CFG.LLM_ASSISTANT_JARGON,
    ui_summarize_commands=["/compress", "/compact"],
    ui_attach_commands=["/attach"],
    ui_exit_commands=["/q", "/bye", "/quit", "/exit"],
    ui_info_commands=["/info", "/help"],
    ui_save_commands=["/save"],
    ui_load_commands=["/load"],
    ui_yolo_toggle_commands=["/yolo"],
    ui_redirect_output_commands=[">", "/redirect"],
    ui_exec_commands=["!", "/exec"],
)
llm_group.add_task(llm_chat)
cli.add_task(llm_chat)


llm_chat.prompt_manager.add_middleware(
    new_prompt(lambda: get_persona_prompt(CFG.LLM_ASSISTANT_NAME)),
    new_prompt(lambda: get_mandate_prompt()),
    system_context,
    create_note_prompt(note_manager),
    create_claude_compatibility_prompt(skill_manager),
    create_zrb_prompt(),
)
llm_chat.add_confirmation_middleware(replace_confirmation)
llm_chat.add_toolset(*load_mcp_config())
llm_chat.add_tool(
    run_shell_command,
    list_files,
    glob_files,
    read_file,
    read_files,
    write_file,
    write_files,
    replace_in_file,
    search_files,
    analyze_file,
    analyze_code,
    search_internet,
    open_web_page,
    create_list_zrb_task_tool(),
    create_run_zrb_task_tool(),
    create_activate_skill_tool(skill_manager),
    *create_note_tools(note_manager),
)



import os
from typing import Callable, List

from zrb.llm.note.manager import NoteManager


def create_note_tools(note_manager: NoteManager) -> List[Callable]:
    async def read_long_term_note() -> str:
        """
        Retrieves your GLOBAL 🧠 Long-Term Memory.
        This contains established preferences, personal facts, and context spanning multiple projects.
        ALWAYS check this at the start of a session.
        """
        return note_manager.read("~")

    read_long_term_note.__name__ = "read_long_term_note"

    async def write_long_term_note(content: str) -> str:
        """
        Updates your GLOBAL 🧠 Long-Term Memory with CRITICAL information.
        Use this to persist user preferences, personal facts, and cross-project rules.

        **WARNING:** This COMPLETELY OVERWRITES the existing Long-Term Note.

        **ARGS:**
        - `content`: The full text to store in the global memory.
        """
        note_manager.write("~", content)
        return "Global long-term note saved."

    write_long_term_note.__name__ = "write_long_term_note"

    async def read_contextual_note(path: str | None = None) -> str:
        """
        Retrieves LOCAL 📝 Contextual Notes for a specific project or directory.
        Use this to recall architectural decisions or project-specific guidelines.

        **ARGS:**
        - `path`: Target file/dir path. Defaults to current working directory.
        """
        if path is None:
            path = os.getcwd()
        return note_manager.read(path)

    read_contextual_note.__name__ = "read_contextual_note"

    async def write_contextual_note(content: str, path: str | None = None) -> str:
        """
        Persists LOCAL 📝 Contextual Notes for a specific project or directory.
        Use this to save architectural patterns or progress markers for the current task.

        **WARNING:** This COMPLETELY OVERWRITES the contextual note for the specified path.

        **ARGS:**
        - `content`: The full text to store in the local memory.
        - `path`: Target file/dir path. Defaults to current working directory.
        """
        if path is None:
            path = os.getcwd()
        note_manager.write(path, content)
        return f"Contextual note saved for: {path}"

    write_contextual_note.__name__ = "write_contextual_note"

    return [
        read_long_term_note,
        write_long_term_note,
        read_contextual_note,
        write_contextual_note,
    ]



from zrb.llm.app.confirmation.allow_tool import allow_tool_usage
from zrb.llm.app.confirmation.handler import ConfirmationMiddleware, last_confirmation
from zrb.llm.app.confirmation.replace_confirmation import replace_confirmation

__all__ = [
    "allow_tool_usage",
    "ConfirmationMiddleware",
    "last_confirmation",
    "replace_confirmation",
]



import json
import re
from typing import Any, Awaitable, Callable, Dict, Optional

from zrb.llm.app.confirmation.handler import ConfirmationMiddleware, UIProtocol


def allow_tool_usage(
    tool_name: str, kwargs: Optional[Dict[str, str]] = None
) -> ConfirmationMiddleware:
    """
    Creates a confirmation middleware that automatically approves a tool execution
    if it matches the specified tool_name and argument constraints.

    :param tool_name: The name of the tool to allow.
    :param kwargs: A dictionary of regex patterns for arguments.
                   If None or empty, the tool is allowed regardless of arguments.
                   If provided, arguments in the tool call must match the regex patterns
                   specified in kwargs (only for arguments present in both).
    :return: A ConfirmationMiddleware function.
    """
    from pydantic_ai import ToolApproved

    async def middleware(
        ui: UIProtocol,
        call: Any,
        response: str,
        next_handler: Callable[[UIProtocol, Any, str], Awaitable[Any]],
    ) -> Any:
        # Check if tool name matches
        if call.tool_name != tool_name:
            return await next_handler(ui, call, response)

        # If kwargs is empty or None, approve
        if not kwargs:
            ui.append_to_output(f"\n✅ Auto-approved tool: {tool_name}")
            return ToolApproved()

        # Parse arguments
        try:
            args = call.args
            if isinstance(args, str):
                args = json.loads(args)

            if not isinstance(args, dict):
                # If args is not a dict (e.g. primitive), and kwargs is not empty,
                # we assume it doesn't match complex constraints (or we can't check keys).
                # So we delegate to the next handler.
                return await next_handler(ui, call, response)

        except (json.JSONDecodeError, ValueError):
            return await next_handler(ui, call, response)

        # Check constraints
        # "all parameter in the call parameter has to match the ones in kwargs (if that parameter defined in the kwargs)"
        for arg_name, arg_value in args.items():
            if arg_name in kwargs:
                pattern = kwargs[arg_name]
                # Convert arg_value to string for regex matching
                if not re.search(pattern, str(arg_value)):
                    return await next_handler(ui, call, response)

        ui.append_to_output(f"\n✅ Auto-approved tool: {tool_name} with matching args")
        return ToolApproved()

    return middleware



import asyncio
import contextlib
import os
import re
from collections.abc import Callable
from datetime import datetime
from typing import TYPE_CHECKING, Any, TextIO

from prompt_toolkit import Application
from prompt_toolkit.application import get_app
from prompt_toolkit.document import Document
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import Layout
from prompt_toolkit.lexers import Lexer
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.styles import Style

from zrb.context.shared_context import SharedContext
from zrb.llm.app.confirmation.handler import (
    ConfirmationHandler,
    ConfirmationMiddleware,
    last_confirmation,
)
from zrb.llm.app.keybinding import create_output_keybindings
from zrb.llm.app.layout import create_input_field, create_layout, create_output_field
from zrb.llm.app.redirection import StreamToUI
from zrb.llm.app.style import create_style
from zrb.llm.history_manager.any_history_manager import AnyHistoryManager
from zrb.llm.task.llm_task import LLMTask
from zrb.session.any_session import AnySession
from zrb.session.session import Session
from zrb.task.any_task import AnyTask
from zrb.util.ascii_art.banner import create_banner
from zrb.util.cli.markdown import render_markdown
from zrb.util.cli.style import stylize_faint
from zrb.util.string.name import get_random_name

if TYPE_CHECKING:
    from pydantic_ai import UserContent
    from pydantic_ai.models import Model
    from rich.theme import Theme


class UI:
    def __init__(
        self,
        greeting: str,
        assistant_name: str,
        ascii_art: str,
        jargon: str,
        output_lexer: Lexer,
        llm_task: LLMTask,
        history_manager: AnyHistoryManager,
        initial_message: Any = "",
        initial_attachments: "list[UserContent]" = [],
        conversation_session_name: str = "",
        yolo: bool = False,
        triggers: list[Callable[[], Any]] = [],
        confirmation_middlewares: list[ConfirmationMiddleware] = [],
        markdown_theme: "Theme | None" = None,
        summarize_commands: list[str] = [],
        attach_commands: list[str] = [],
        exit_commands: list[str] = [],
        info_commands: list[str] = [],
        save_commands: list[str] = [],
        load_commands: list[str] = [],
        redirect_output_commands: list[str] = [],
        yolo_toggle_commands: list[str] = [],
        exec_commands: list[str] = [],
        model: "Model | str | None" = None,
    ):
        self._is_thinking = False
        self._running_llm_task: asyncio.Task | None = None
        self._llm_task = llm_task
        self._history_manager = history_manager
        self._assistant_name = assistant_name
        self._ascii_art = ascii_art
        self._jargon = jargon
        self._initial_message = initial_message
        self._conversation_session_name = conversation_session_name
        if not self._conversation_session_name:
            self._conversation_session_name = get_random_name()
        self._yolo = yolo
        self._model = model
        self._triggers = triggers
        self._markdown_theme = markdown_theme
        self._summarize_commands = summarize_commands
        self._attach_commands = attach_commands
        self._exit_commands = exit_commands
        self._info_commands = info_commands
        self._save_commands = save_commands
        self._load_commands = load_commands
        self._redirect_output_commands = redirect_output_commands
        self._yolo_toggle_commands = yolo_toggle_commands
        self._exec_commands = exec_commands
        self._trigger_tasks: list[asyncio.Task] = []
        self._last_result_data: str | None = None
        # System Info
        self._cwd = os.getcwd()
        self._git_info = "Checking..."
        self._system_info_task: asyncio.Task | None = None
        # Attachments
        self._pending_attachments: "list[UserContent]" = list(initial_attachments)
        # Confirmation Handler
        self._confirmation_handler = ConfirmationHandler(
            middlewares=confirmation_middlewares + [last_confirmation]
        )
        # Confirmation state (Used by ask_user and keybindings)
        self._waiting_for_confirmation = False
        self._confirmation_future: asyncio.Future[str] | None = None
        # UI Styles
        self._style = create_style()
        # Input Area
        self._input_field = create_input_field(
            history_manager=self._history_manager,
            attach_commands=self._attach_commands,
            exit_commands=self._exit_commands,
            info_commands=self._info_commands,
            save_commands=self._save_commands,
            load_commands=self._load_commands,
            redirect_output_commands=self._redirect_output_commands,
            summarize_commands=self._summarize_commands,
            exec_commands=self._exec_commands,
        )
        # Output Area (Read-only chat history)
        help_text = self._get_help_text()
        full_greeting = create_banner(self._ascii_art, f"{greeting}\n{help_text}")
        self._output_field = create_output_field(full_greeting, output_lexer)
        self._output_field.control.key_bindings = create_output_keybindings(
            self._input_field
        )
        self._layout = create_layout(
            title=self._assistant_name,
            jargon=self._jargon,
            input_field=self._input_field,
            output_field=self._output_field,
            info_bar_text=self._get_info_bar_text,
            status_bar_text=self._get_status_bar_text,
        )
        # Key Bindings
        self._app_kb = KeyBindings()
        self._setup_app_keybindings(
            app_keybindings=self._app_kb, llm_task=self._llm_task
        )
        # Application
        self._application = self._create_application(
            layout=self._layout, keybindings=self._app_kb, style=self._style
        )
        # Send message if first_message is provided. Make sure only run at most once
        if self._initial_message:
            self._application.after_render.add_handler(self._on_first_render)

    async def run_async(self):
        """Run the application and manage triggers."""
        import logging

        # Start triggers
        for trigger_fn in self._triggers:
            trigger_task = self._application.create_background_task(
                self._trigger_loop(trigger_fn)
            )
            self._trigger_tasks.append(trigger_task)

        # Start system info update loop
        self._system_info_task = self._application.create_background_task(
            self._update_system_info_loop()
        )

        # Setup logging redirection to UI
        root_logger = logging.getLogger()
        original_handlers = root_logger.handlers[:]

        ui_stream = StreamToUI(self.append_to_output)
        ui_handler = logging.StreamHandler(ui_stream)
        formatter = logging.Formatter("[%(levelname)s] %(name)s: %(message)s")
        ui_handler.setFormatter(formatter)

        root_logger.handlers = [ui_handler]

        try:
            with patch_stdout():
                return await self._application.run_async()
        finally:
            # Restore handlers
            root_logger.handlers = original_handlers

            # Stop triggers
            for trigger_task in self._trigger_tasks:
                trigger_task.cancel()
            self._trigger_tasks.clear()

            if self._system_info_task:
                self._system_info_task.cancel()

    async def _trigger_loop(self, trigger_fn: Callable[[], Any]):
        """Handle external triggers and submit user message when trigger activated"""
        while True:
            try:
                stdout_capture = StreamToUI(self.append_to_output)
                with contextlib.redirect_stdout(
                    stdout_capture
                ), contextlib.redirect_stderr(stdout_capture):
                    if asyncio.iscoroutinefunction(trigger_fn):
                        result = await trigger_fn()
                    else:
                        result = await asyncio.to_thread(trigger_fn)

                if result:
                    self._submit_user_message(self._llm_task, str(result))

            except asyncio.CancelledError:
                break
            except Exception:
                # Keep running on error, maybe log it
                pass

    async def _update_system_info_loop(self):
        """Periodically update CWD and Git info."""
        while True:
            try:
                self._cwd = self._get_cwd_display()
                branch, status = await self._get_git_info()
                if branch:
                    self._git_info = f"{branch}{status}"
                else:
                    self._git_info = "Not a git repo"
                get_app().invalidate()
            except asyncio.CancelledError:
                break
            except Exception:
                pass
            await asyncio.sleep(2)

    def _get_cwd_display(self) -> str:
        cwd = os.getcwd()
        home = os.path.expanduser("~")
        if cwd.startswith(home):
            return "~" + cwd[len(home) :]
        return cwd

    async def _get_git_info(self) -> tuple[str, str]:
        """Returns (branch_name, status_symbol)"""
        try:
            # Check branch
            proc = await asyncio.create_subprocess_exec(
                "git",
                "rev-parse",
                "--abbrev-ref",
                "HEAD",
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, _ = await proc.communicate()
            if proc.returncode != 0:
                return "", ""
            branch = stdout.decode().strip()

            # Check status (dirty or clean)
            proc = await asyncio.create_subprocess_exec(
                "git",
                "status",
                "--porcelain",
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, _ = await proc.communicate()
            is_dirty = bool(stdout.strip())

            return branch, "*" if is_dirty else ""
        except Exception:
            return "", ""

    def _on_first_render(self, app: Application):
        """Handle initial message (the message sent when creating the UI)"""
        self._application.after_render.remove_handler(self._on_first_render)
        self._submit_user_message(self._llm_task, self._initial_message)

    async def ask_user(self, prompt: str) -> str:
        """Prompts the user for input via the main input field, blocking until provided."""
        self._waiting_for_confirmation = True
        self._confirmation_future = asyncio.Future()

        if prompt:
            self.append_to_output(prompt, end="")

        get_app().invalidate()

        try:
            return await self._confirmation_future
        finally:
            self._waiting_for_confirmation = False
            self._confirmation_future = None

    async def _confirm_tool_execution(self, call: Any) -> Any:
        return await self._confirmation_handler.handle(self, call)

    @property
    def triggers(self) -> list[Callable[[], Any]]:
        return self._triggers

    @triggers.setter
    def triggers(self, value: list[Callable[[], Any]]):
        self._triggers = value

    @property
    def application(self) -> Application:
        return self._application

    @property
    def last_output(self) -> str:
        if self._last_result_data is None:
            return ""
        return self._last_result_data

    def _create_application(
        self,
        layout: Layout,
        keybindings: KeyBindings,
        style: Style,
    ) -> Application:
        return Application(
            layout=layout,
            key_bindings=keybindings,
            style=style,
            full_screen=True,
            mouse_support=True,
        )

    def _get_help_text(self) -> str:
        help_lines = ["\nAvailable Commands:"]

        def add_cmd_help(commands: list[str], description: str):
            if commands and len(commands) > 0:
                cmd = commands[0]
                description = description.replace("{cmd}", cmd)
                help_lines.append(f"  {cmd:<10} : {description}")

        add_cmd_help(self._exit_commands, "Exit the application")
        add_cmd_help(self._info_commands, "Show this help message")
        add_cmd_help(self._attach_commands, "Attach file (usage: {cmd} <path>)")
        add_cmd_help(self._save_commands, "Save conversation (usage: {cmd} <name>)")
        add_cmd_help(self._load_commands, "Load conversation (usage: {cmd} <name>)")
        add_cmd_help(
            self._redirect_output_commands,
            "Save last output to file (usage: {cmd} <file>)",
        )
        add_cmd_help(self._summarize_commands, "Summarize conversation history")
        add_cmd_help(self._yolo_toggle_commands, "Toggle YOLO mode")
        add_cmd_help(
            self._exec_commands, "Execute shell command (usage: {cmd} <command>)"
        )

        return "\n".join(help_lines) + "\n"

    def _get_info_bar_text(self) -> AnyFormattedText:
        from prompt_toolkit.formatted_text import HTML

        model_name = "Unknown"
        if self._model:
            if isinstance(self._model, str):
                model_name = self._model
            elif hasattr(self._model, "model_name"):
                model_name = getattr(self._model, "model_name")
            else:
                model_name = str(self._model)
        yolo_text = (
            "<style color='ansired'><b>ON</b></style>"
            if self._yolo
            else "<style color='ansigreen'>OFF</style>"
        )
        return HTML(
            f" 🤖 <b>Model:</b> {model_name} "
            f"| 🗣️ <b>Session:</b> {self._conversation_session_name} "
            f"| 🤠 <b>YOLO:</b> {yolo_text} \n"
            f" 📂 <b>Dir:</b> {self._cwd} "
            f"| 🌿 <b>Git:</b> {self._git_info} "
        )

    def _get_status_bar_text(self) -> AnyFormattedText:
        if self._is_thinking:
            return [("class:thinking", f" {self._assistant_name} is thinking... ")]
        return [("class:status", " Ready ")]

    def _setup_app_keybindings(self, app_keybindings: KeyBindings, llm_task: AnyTask):
        @app_keybindings.add("c-c")
        def _(event):
            # If text is selected, copy it instead of exiting
            buffer = event.app.current_buffer
            if buffer.selection_state:
                data = buffer.copy_selection()
                event.app.clipboard.set_data(data)
                buffer.exit_selection()
                return
            # If buffer is not empty, clear it
            if buffer.text != "":
                buffer.reset()
                return
            event.app.exit()

        @app_keybindings.add("escape")
        def _(event):
            if self._running_llm_task and not self._running_llm_task.done():
                self._running_llm_task.cancel()
                self.append_to_output("\n<Esc> Canceled")

        @app_keybindings.add("enter")
        def _(event):
            # Handle confirmation and multiline
            if self._handle_confirmation(event):
                return
            if self._handle_multiline(event):
                return

            # Handle empty inputs
            buff = event.current_buffer
            text = buff.text
            if not text.strip():
                return

            # Handle other commands
            if self._handle_exit_command(event):
                return
            if self._handle_info_command(event):
                return
            if self._handle_save_command(event):
                return
            if self._handle_load_command(event):
                return
            if self._handle_redirect_command(event):
                return
            if self._handle_attach_command(event):
                return
            if self._handle_toggle_yolo(event):
                return
            if self._handle_exec_command(event):
                return

            # If we are thinking, ignore input
            if self._is_thinking:
                return
            self._submit_user_message(llm_task, text)
            buff.reset()

        @app_keybindings.add("c-j")  # Ctrl+J
        @app_keybindings.add("c-space")  # Ctrl+Space (Fallback)
        def _(event):
            event.current_buffer.insert_text("\n")

    def _handle_exec_command(self, event) -> bool:
        buff = event.current_buffer
        text = buff.text
        for cmd in self._exec_commands:
            prefix = f"{cmd} "
            if text.strip().lower().startswith(prefix):
                if self._is_thinking:
                    return False

                shell_cmd = text.strip()[len(prefix) :].strip()
                if not shell_cmd:
                    return True

                buff.reset()
                # Run in background
                self._running_llm_task = asyncio.create_task(
                    self._run_shell_command(shell_cmd)
                )
                return True
        return False

    async def _run_shell_command(self, cmd: str):
        self._is_thinking = True
        get_app().invalidate()
        timestamp = datetime.now().strftime("%H:%M")

        try:
            self.append_to_output(f"\n💻 {timestamp} >>\n$ {cmd}\n")
            self.append_to_output(f"\n  🔢 Executing...\n")

            # Create subprocess
            process = await asyncio.create_subprocess_shell(
                cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            is_first_output = True

            # Read output streams
            async def read_stream(stream, is_stderr=False):
                while True:
                    nonlocal is_first_output
                    line = await stream.readline()
                    if not line:
                        break
                    decoded_line = line.decode("utf-8", errors="replace")
                    decoded_line = decoded_line.replace("\n", "\n  ").replace(
                        "\r", "\r  "
                    )
                    if is_first_output:
                        decoded_line = f"  {decoded_line}"
                        is_first_output = False
                    # Could use a different color for stderr if desired
                    self.append_to_output(decoded_line, end="")

            await asyncio.gather(
                read_stream(process.stdout), read_stream(process.stderr, is_stderr=True)
            )

            return_code = await process.wait()

            if return_code == 0:
                self.append_to_output(f"\n  ✅ Command finished successfully.\n")
            else:
                self.append_to_output(
                    f"\n  ❌ Command failed with exit code {return_code}.\n"
                )

        except asyncio.CancelledError:
            self.append_to_output("\n[Cancelled]\n")
        except Exception as e:
            self.append_to_output(f"\n[Error: {e}]\n")
        finally:
            self._is_thinking = False
            self._running_llm_task = None
            get_app().invalidate()

    def _handle_toggle_yolo(self, event) -> bool:
        buff = event.current_buffer
        text = buff.text
        if text.strip().lower() in self._yolo_toggle_commands:
            if self._is_thinking:
                return False
            self._yolo = not self._yolo
            buff.reset()
            return True
        return False

    def _handle_multiline(self, event) -> bool:
        buff = event.current_buffer
        text = buff.text
        # Check for multiline indicator (trailing backslash)
        if text.strip().endswith("\\"):
            # If cursor is at the end, remove backslash and insert newline
            if buff.cursor_position == len(text):
                # Remove the backslash (assuming it's the last char)
                # We need to be careful with whitespace after backslash if we used strip()
                # Let's just check the character before cursor
                if text.endswith("\\"):
                    buff.delete_before_cursor(count=1)
                    buff.insert_text("\n")
                    return True
        return False

    def _handle_confirmation(self, event) -> bool:
        buff = event.current_buffer
        text = buff.text
        if self._waiting_for_confirmation and self._confirmation_future:
            # Echo the user input
            self.append_to_output(text + "\n")
            if not self._confirmation_future.done():
                self._confirmation_future.set_result(text)
            buff.reset()
            return True
        return False

    def _handle_exit_command(self, event) -> bool:
        buff = event.current_buffer
        text = buff.text
        if text.strip().lower() in self._exit_commands:
            event.app.exit()
            return True
        return False

    def _handle_info_command(self, event) -> bool:
        buff = event.current_buffer
        text = buff.text
        if text.strip().lower() in self._info_commands:
            self.append_to_output(stylize_faint(self._get_help_text()))
            buff.reset()
            return True
        return False

    def _handle_save_command(self, event) -> bool:
        buff = event.current_buffer
        text = buff.text.strip()
        for cmd in self._save_commands:
            prefix = f"{cmd} "
            if text.lower().startswith(prefix):
                name = text[len(prefix) :].strip()
                if not name:
                    continue
                try:
                    history = self._history_manager.load(
                        self._conversation_session_name
                    )
                    self._history_manager.update(name, history)
                    self._history_manager.save(name)
                    self.append_to_output(f"\n  💾 Conversation saved as: {name}\n")
                except Exception as e:
                    self.append_to_output(f"\n  ❌ Failed to save conversation: {e}\n")
                buff.reset()
                return True
        return False

    def _handle_load_command(self, event) -> bool:
        buff = event.current_buffer
        text = buff.text.strip()
        for cmd in self._load_commands:
            prefix = f"{cmd} "
            if text.lower().startswith(prefix):
                name = text[len(prefix) :].strip()
                if not name:
                    continue
                self._conversation_session_name = name
                self.append_to_output(
                    f"\n  📂 Conversation session switched to: {name}\n"
                )
                buff.reset()
                return True
        return False

    def _handle_redirect_command(self, event) -> bool:
        buff = event.current_buffer
        text = buff.text.strip()
        for cmd in self._redirect_output_commands:
            prefix = f"{cmd} "
            if text.lower().startswith(prefix):
                path = text[len(prefix) :].strip()
                if not path:
                    continue

                content = self.last_output
                if not content:
                    self.append_to_output(
                        "\n  ❌ No AI response available to redirect.\n"
                    )
                    buff.reset()
                    return True

                try:
                    # Write to file
                    expanded_path = os.path.abspath(os.path.expanduser(path))
                    os.makedirs(os.path.dirname(expanded_path), exist_ok=True)
                    with open(expanded_path, "w", encoding="utf-8") as f:
                        f.write(content)
                    self.append_to_output(f"\n  📝 Last output redirected to: {path}\n")
                except Exception as e:
                    self.append_to_output(f"\n  ❌ Failed to redirect output: {e}\n")

                buff.reset()
                return True
        return False

    def _handle_attach_command(self, event) -> bool:
        buff = event.current_buffer
        text = buff.text
        for attach_command in self._attach_commands:
            if text.strip().lower().startswith(f"{attach_command} "):
                path = text.strip()[8:].strip()
                self._submit_attachment(path)
                buff.reset()
                return True
        return False

    def _submit_attachment(self, path: str):
        # Validate path
        self.append_to_output(f"\n  🔢 Attach {path}...\n")
        expanded_path = os.path.abspath(os.path.expanduser(path))
        if not os.path.exists(expanded_path):
            self.append_to_output(f"\n  ❌ File not found: {path}\n")
            return
        if expanded_path not in self._pending_attachments:
            self._pending_attachments.append(expanded_path)
            self.append_to_output(f"\n  📎 Attached: {path}\n")
        else:
            self.append_to_output(f"\n  📎 Already attached: {path}\n")

    def append_to_output(
        self,
        *values: object,
        sep: str = " ",
        end: str = "\n",
        file: TextIO | None = None,
        flush: bool = False,
    ):
        # Helper to safely append to read-only buffer
        current_text = self._output_field.text

        # Construct the new content
        content = sep.join([str(value) for value in values]) + end

        # Handle carriage returns (\r) for status updates
        if "\r" in content:
            # Find the start of the last line in the current text
            last_newline = current_text.rfind("\n")
            if last_newline == -1:
                previous = ""
                last = current_text
            else:
                previous = current_text[: last_newline + 1]
                last = current_text[last_newline + 1 :]

            combined = last + content
            # Remove content before \r on the same line
            # [^\n]* matches any character except newline
            resolved = re.sub(r"[^\n]*\r", "", combined)

            new_text = previous + resolved
        else:
            new_text = current_text + content

        # Update content directly
        # We use bypass_readonly=True by constructing a Document
        self._output_field.buffer.set_document(
            Document(new_text, cursor_position=len(new_text)), bypass_readonly=True
        )
        get_app().invalidate()

    def _submit_user_message(self, llm_task: AnyTask, user_message: str):
        timestamp = datetime.now().strftime("%H:%M")
        # 1. Render User Message
        self.append_to_output(f"\n💬 {timestamp} >>\n{user_message.strip()}\n")
        # 2. Trigger AI Response
        attachments = list(self._pending_attachments)
        self._pending_attachments.clear()
        self._running_llm_task = asyncio.create_task(
            self._stream_ai_response(llm_task, user_message, attachments)
        )

    async def _stream_ai_response(
        self,
        llm_task: AnyTask,
        user_message: str,
        attachments: "list[UserContent]" = [],
    ):
        from zrb.llm.agent.agent import tool_confirmation_var

        self._is_thinking = True
        get_app().invalidate()  # Update status bar
        try:
            timestamp = datetime.now().strftime("%H:%M")
            # Header first
            self.append_to_output(f"\n🤖 {timestamp} >>\n")
            session = self._create_sesion_for_llm_task(user_message, attachments)

            # Run the task with stdout/stderr redirected to UI
            self.append_to_output("\n  🔢 Streaming response...\n")
            stdout_capture = StreamToUI(self.append_to_output)

            # Set context var for tool confirmation
            token = tool_confirmation_var.set(self._confirm_tool_execution)
            try:
                with contextlib.redirect_stdout(
                    stdout_capture
                ), contextlib.redirect_stderr(stdout_capture):
                    result_data = await llm_task.async_run(session)
            finally:
                tool_confirmation_var.reset(token)

            # Check for final text output
            if result_data is not None:
                if isinstance(result_data, str):
                    self._last_result_data = result_data
                    width = self._get_output_field_width()
                    self.append_to_output("\n")
                    self.append_to_output(
                        render_markdown(
                            result_data, width=width, theme=self._markdown_theme
                        )
                    )

        except asyncio.CancelledError:
            self.append_to_output("\n[Cancelled]\n")
        except Exception as e:
            self.append_to_output(f"\n[Error: {e}]\n")
        finally:
            self._is_thinking = False
            self._running_llm_task = None
            get_app().invalidate()

    def _create_sesion_for_llm_task(
        self, user_message: str, attachments: "list[UserContent]"
    ) -> AnySession:
        """Create session to run LLMTask"""
        session_input = {
            "message": user_message,
            "session": self._conversation_session_name,
            "yolo": self._yolo,
            "attachments": attachments,
        }
        shared_ctx = SharedContext(
            input=session_input,
            print_fn=self.append_to_output,
            is_web_mode=True,
        )
        return Session(shared_ctx)

    def _get_output_field_width(self) -> int | None:
        try:
            width = get_app().output.get_size().columns - 4
            if width < 10:
                width = None
        except Exception:
            width = None
        return width



from prompt_toolkit.styles import Style


def create_style() -> Style:
    return Style.from_dict(
        {
            "frame.label": "bg:#000000 #ffff00",
            "thinking": "ansigreen italic",
            "faint": "#888888",
            "output_field": "bg:#000000 #eeeeee",
            "input_field": "bg:#000000 #eeeeee",
            "text": "#eeeeee",
            "status": "reverse",
            "bottom-toolbar": "bg:#333333 #aaaaaa",
        }
    )



from typing import Any, Callable

from zrb.context.any_context import AnyContext
from zrb.context.any_shared_context import AnySharedContext

fstring = str
AnyAttr = Any | fstring | Callable[[AnyContext | AnySharedContext], Any]
StrAttr = str | fstring | Callable[[AnyContext | AnySharedContext], str | None]
BoolAttr = bool | fstring | Callable[[AnyContext | AnySharedContext], bool | None]
IntAttr = int | fstring | Callable[[AnyContext | AnySharedContext], int | None]
FloatAttr = float | fstring | Callable[[AnyContext | AnySharedContext], float | None]
StrDictAttr = (
    dict[str, StrAttr] | Callable[[AnyContext | AnySharedContext], dict[str, Any]]
)
StrListAttr = list[StrAttr] | Callable[[AnyContext | AnySharedContext], list[str]]



from typing import TYPE_CHECKING

from zrb.llm.agent.agent import create_agent
from zrb.llm.prompt.default import get_summarizer_system_prompt

if TYPE_CHECKING:
    from pydantic_ai import Agent
    from pydantic_ai.models import Model


def create_summarizer_agent(
    model: "str | None | Model" = None,
    system_prompt: str | None = None,
) -> "Agent[None, str]":
    effective_system_prompt = system_prompt or get_summarizer_system_prompt()

    return create_agent(
        model=model,
        system_prompt=effective_system_prompt,
    )



import os
from typing import Annotated

from fastapi import Depends, FastAPI
from my_app_name.common.error import ForbiddenError
from my_app_name.module.gateway.util.auth import get_current_user
from my_app_name.module.gateway.util.view import render_content, render_error
from my_app_name.schema.user import AuthUserResponse


def serve_my_module_route(app: FastAPI):
    """
    Serving routes for my_module
    """



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pydantic_ai import ModelMessage


class AnyHistoryManager(ABC):
    @abstractmethod
    def load(self, conversation_name: str) -> "list[ModelMessage]":
        pass

    @abstractmethod
    def save(self, conversation_name: str):
        pass

    @abstractmethod
    def update(self, conversation_name: str, messages: "list[ModelMessage]"):
        pass

    @abstractmethod
    def search(self, keyword: str) -> list[str]:
        pass



from typing import TYPE_CHECKING, Any, Awaitable, Callable

from zrb.config.config import CFG
from zrb.llm.agent.summarizer import create_summarizer_agent
from zrb.llm.config.limiter import LLMLimiter
from zrb.llm.config.limiter import llm_limiter as default_llm_limiter
from zrb.util.markdown import make_markdown_section

if TYPE_CHECKING:
    from pydantic_ai.messages import ModelMessage
else:
    ModelMessage = Any


def is_turn_start(msg: Any) -> bool:
    """Identify start of a new user interaction (User Prompt without Tool Return)."""
    from pydantic_ai.messages import ModelRequest, ToolReturnPart, UserPromptPart

    if not isinstance(msg, ModelRequest):
        return False
    # In pydantic_ai, ModelRequest parts can be list of various parts
    has_user = any(isinstance(p, UserPromptPart) for p in msg.parts)
    has_return = any(isinstance(p, ToolReturnPart) for p in msg.parts)
    return has_user and not has_return


async def summarize_history(
    messages: "list[ModelMessage]",
    agent: Any = None,
    summary_window: int | None = None,
) -> "list[ModelMessage]":
    """
    Summarizes the history, keeping the last `summary_window` messages intact.
    Returns a new list of messages where older messages are replaced by a summary.
    """
    from pydantic_ai.messages import ModelRequest, UserPromptPart

    if summary_window is None:
        summary_window = CFG.LLM_HISTORY_SUMMARIZATION_WINDOW
    if len(messages) <= summary_window:
        return messages

    # Determine split index
    # We want to keep at least summary_window messages.
    # So split_idx <= len(messages) - summary_window.
    # We search backwards from there for a clean turn start.
    start_search_idx = max(0, len(messages) - summary_window)
    split_idx = -1

    # Iterate backwards from start_search_idx to 0
    for i in range(start_search_idx, -1, -1):
        if is_turn_start(messages[i]):
            split_idx = i
            break

    if split_idx <= 0:
        return messages

    to_summarize = messages[:split_idx]
    to_keep = messages[split_idx:]

    # Simple text representation for now
    history_text = "\n".join([str(m) for m in to_summarize])

    summarizer_agent = agent or create_summarizer_agent()

    # Run the summarizer agent
    result = await summarizer_agent.run(
        f"Summarize this conversation history:\n{history_text}"
    )
    summary_text = result.output

    # Create a summary message injected as user context
    summary_message = ModelRequest(
        parts=[
            UserPromptPart(
                content=make_markdown_section(
                    "Previous conversation summary", summary_text
                )
            )
        ]
    )

    return [summary_message] + to_keep


def create_summarizer_history_processor(
    agent: Any = None,
    limiter: LLMLimiter | None = None,
    token_threshold: int | None = None,
    summary_window: int | None = None,
) -> "Callable[[list[ModelMessage]], Awaitable[list[ModelMessage]]]":
    """
    Creates a history processor that auto-summarizes history when it exceeds `token_threshold`.
    """
    llm_limiter = limiter or default_llm_limiter
    if token_threshold is None:
        token_threshold = CFG.LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD

    async def process_history(messages: "list[ModelMessage]") -> "list[ModelMessage]":
        current_tokens = llm_limiter.count_tokens(messages)

        if current_tokens <= token_threshold:
            return messages

        return await summarize_history(messages, agent, summary_window)

    return process_history



import os
from typing import TYPE_CHECKING

from zrb.config.web_auth_config import WebAuthConfig
from zrb.util.file import read_file

if TYPE_CHECKING:
    # We want fastapi to only be loaded when necessary to decrease footprint
    from fastapi import FastAPI


def serve_static_resources(app: "FastAPI", web_auth_config: WebAuthConfig) -> None:
    from fastapi import HTTPException
    from fastapi.responses import FileResponse, PlainTextResponse
    from fastapi.staticfiles import StaticFiles

    _STATIC_DIR = os.path.join(os.path.dirname(__file__), "resources")

    app.mount("/static", StaticFiles(directory=_STATIC_DIR), name="static")

    # Serve static files
    @app.get("/static/{file_path:path}", include_in_schema=False)
    async def static_files(file_path: str):
        full_path = os.path.join(_STATIC_DIR, file_path)
        if os.path.isfile(full_path):
            return FileResponse(full_path)
        raise HTTPException(status_code=404, detail="File not found")

    @app.get("/refresh-token.js", include_in_schema=False)
    async def refresh_token_js():
        return PlainTextResponse(
            content=_get_refresh_token_js(
                60 * web_auth_config.access_token_expire_minutes / 3
            ),
            media_type="application/javascript",
        )


def _get_refresh_token_js(refresh_interval_seconds: int):
    _DIR = os.path.dirname(__file__)
    return read_file(
        os.path.join(_DIR, "refresh-token.template.js"),
        {"refreshIntervalSeconds": f"{refresh_interval_seconds}"},
    )



from logging import Logger

from my_app_name.common.base_service import BaseService
from my_app_name.module.auth.service.permission.repository.permission_repository import (
    PermissionRepository,
)
from my_app_name.schema.permission import (
    MultiplePermissionResponse,
    PermissionCreateWithAudit,
    PermissionResponse,
    PermissionUpdateWithAudit,
)


class PermissionService(BaseService):

    def __init__(self, logger: Logger, permission_repository: PermissionRepository):
        super().__init__(logger)
        self.permission_repository = permission_repository

    @BaseService.route(
        "/api/v1/permissions/{permission_id}",
        methods=["get"],
        response_model=PermissionResponse,
    )
    async def get_permission_by_id(self, permission_id: str) -> PermissionResponse:
        return await self.permission_repository.get_by_id(permission_id)

    @BaseService.route(
        "/api/v1/permissions",
        methods=["get"],
        response_model=MultiplePermissionResponse,
    )
    async def get_permissions(
        self,
        page: int = 1,
        page_size: int = 10,
        sort: str | None = None,
        filter: str | None = None,
    ) -> MultiplePermissionResponse:
        permissions = await self.permission_repository.get(
            page, page_size, filter, sort
        )
        count = await self.permission_repository.count(filter)
        return MultiplePermissionResponse(data=permissions, count=count)

    @BaseService.route(
        "/api/v1/permissions/bulk",
        methods=["post"],
        response_model=list[PermissionResponse],
    )
    async def create_permission_bulk(
        self, data: list[PermissionCreateWithAudit]
    ) -> list[PermissionResponse]:
        permissions = await self.permission_repository.create_bulk(data)
        return await self.permission_repository.get_by_ids(
            [permission.id for permission in permissions]
        )

    @BaseService.route(
        "/api/v1/permissions",
        methods=["post"],
        response_model=PermissionResponse,
    )
    async def create_permission(
        self, data: PermissionCreateWithAudit
    ) -> PermissionResponse:
        permission = await self.permission_repository.create(data)
        return await self.permission_repository.get_by_id(permission.id)

    @BaseService.route(
        "/api/v1/permissions/bulk",
        methods=["put"],
        response_model=list[PermissionResponse],
    )
    async def update_permission_bulk(
        self, permission_ids: list[str], data: PermissionUpdateWithAudit
    ) -> list[PermissionResponse]:
        await self.permission_repository.update_bulk(permission_ids, data)
        return await self.permission_repository.get_by_ids(permission_ids)

    @BaseService.route(
        "/api/v1/permissions/{permission_id}",
        methods=["put"],
        response_model=PermissionResponse,
    )
    async def update_permission(
        self, permission_id: str, data: PermissionUpdateWithAudit
    ) -> PermissionResponse:
        await self.permission_repository.update(permission_id, data)
        return await self.permission_repository.get_by_id(permission_id)

    @BaseService.route(
        "/api/v1/permissions/bulk",
        methods=["delete"],
        response_model=list[PermissionResponse],
    )
    async def delete_permission_bulk(
        self, permission_ids: list[str], deleted_by: str
    ) -> list[PermissionResponse]:
        permissions = await self.permission_repository.get_by_ids(permission_ids)
        await self.permission_repository.delete_bulk(permission_ids)
        return permissions

    @BaseService.route(
        "/api/v1/permissions/{permission_id}",
        methods=["delete"],
        response_model=PermissionResponse,
    )
    async def delete_permission(
        self, permission_id: str, deleted_by: str
    ) -> PermissionResponse:
        permission = await self.permission_repository.get_by_id(permission_id)
        await self.permission_repository.delete(permission_id)
        return permission



import os
from datetime import datetime
from typing import Iterable

from prompt_toolkit.completion import (
    CompleteEvent,
    Completer,
    Completion,
    PathCompleter,
)
from prompt_toolkit.document import Document

from zrb.llm.history_manager.any_history_manager import AnyHistoryManager
from zrb.util.match import fuzzy_match


class InputCompleter(Completer):
    def __init__(
        self,
        history_manager: AnyHistoryManager,
        attach_commands: list[str] = [],
        exit_commands: list[str] = [],
        info_commands: list[str] = [],
        save_commands: list[str] = [],
        load_commands: list[str] = [],
        redirect_output_commands: list[str] = [],
        summarize_commands: list[str] = [],
        exec_commands: list[str] = [],
    ):
        self._history_manager = history_manager
        self._attach_commands = attach_commands
        self._exit_commands = exit_commands
        self._info_commands = info_commands
        self._save_commands = save_commands
        self._load_commands = load_commands
        self._redirect_output_commands = redirect_output_commands
        self._summarize_commands = summarize_commands
        self._exec_commands = exec_commands
        # expanduser=True allows ~/path
        self._path_completer = PathCompleter(expanduser=True)
        # Cache for file listing to improve performance
        self._file_cache: list[str] | None = None
        self._file_cache_time = 0
        self._cmd_history = self._get_cmd_history()

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        text_before_cursor = document.text_before_cursor.lstrip()
        word = document.get_word_before_cursor(WORD=True)

        all_commands = (
            self._exit_commands
            + self._attach_commands
            + self._summarize_commands
            + self._info_commands
            + self._save_commands
            + self._load_commands
            + self._redirect_output_commands
            + self._exec_commands
        )
        command_prefixes = {cmd[0] for cmd in all_commands if cmd}

        # 1. Command and Argument Completion
        if text_before_cursor and text_before_cursor[0] in command_prefixes:
            parts = text_before_cursor.split()
            # Check if we are typing the command itself or arguments
            is_typing_command = len(parts) == 1 and not text_before_cursor.endswith(" ")
            is_typing_arg = (len(parts) == 1 and text_before_cursor.endswith(" ")) or (
                len(parts) >= 2
            )

            if is_typing_command:
                lower_word = word.lower()
                prefix = text_before_cursor[0]
                for cmd in all_commands:
                    if cmd.startswith(prefix) and cmd.lower().startswith(lower_word):
                        yield Completion(cmd, start_position=-len(word))
                return

            if is_typing_arg:
                cmd = parts[0]
                arg_prefix = text_before_cursor[len(cmd) :].lstrip()

                # Exec Command: Suggest History
                if self._is_command(cmd, self._exec_commands):
                    # Filter history
                    matches = [h for h in self._cmd_history if h.startswith(arg_prefix)]
                    # Sort matches by length (shorter first) as heuristic? Or just recent?
                    # Since _cmd_history is set (unique), we lose order.
                    # But Python 3.7+ dicts preserve insertion order, so if we used dict keys, we kept order.
                    # Let's assume _get_cmd_history returns recent last.
                    # We reverse to show most recent first.
                    for h in reversed(matches):
                        yield Completion(h, start_position=-len(arg_prefix))
                    return

                # Check if we are typing the second part (argument) strictly
                # (Re-evaluating logic for other commands which only take 1 arg usually)
                if not (
                    (len(parts) == 1 and text_before_cursor.endswith(" "))
                    or (len(parts) == 2 and not text_before_cursor.endswith(" "))
                ):
                    return

                arg_prefix = parts[1] if len(parts) == 2 else ""

                # Save Command: Suggest Timestamp
                if self._is_command(cmd, self._save_commands):
                    ts = datetime.now().strftime("%Y-%m-%d-%H-%M")
                    if ts.startswith(arg_prefix):
                        yield Completion(ts, start_position=-len(arg_prefix))
                    return

                # Redirect Command: Suggest Timestamp.txt
                if self._is_command(cmd, self._redirect_output_commands):
                    ts = datetime.now().strftime("%Y-%m-%d-%H-%M.txt")
                    if ts.startswith(arg_prefix):
                        yield Completion(ts, start_position=-len(arg_prefix))
                    return

                # Load Command: Search History
                if self._is_command(cmd, self._load_commands):
                    results = self._history_manager.search(arg_prefix)
                    for res in results[:10]:
                        yield Completion(res, start_position=-len(arg_prefix))
                    return

                # Attach Command: Suggest Files
                if self._is_command(cmd, self._attach_commands):
                    yield from self._get_file_completions(
                        arg_prefix, complete_event, only_files=True
                    )
                    return

                # Other commands (Exit, Info, Summarize) need no completion
                return

        # 2. File Completion (@)
        if word.startswith("@"):
            path_part = word[1:]
            yield from self._get_file_completions(
                path_part, complete_event, only_files=False
            )

    def _get_cmd_history(self) -> list[str]:
        history_files = [
            os.path.expanduser("~/.bash_history"),
            os.path.expanduser("~/.zsh_history"),
        ]
        unique_cmds = {}  # Use dict to preserve order (insertion order)

        for hist_file in history_files:
            if not os.path.exists(hist_file):
                continue
            try:
                with open(hist_file, "r", errors="ignore") as f:
                    for line in f:
                        line = line.strip()
                        if not line:
                            continue
                        # Handle zsh timestamp format: : 1612345678:0;command
                        if line.startswith(": ") and ";" in line:
                            parts = line.split(";", 1)
                            if len(parts) == 2:
                                line = parts[1]

                        if line:
                            # Remove existing to update position to end (most recent)
                            if line in unique_cmds:
                                del unique_cmds[line]
                            unique_cmds[line] = None
            except Exception:
                pass

        return list(unique_cmds.keys())

    def _is_command(self, cmd: str, cmd_list: list[str]) -> bool:
        return cmd.lower() in [c.lower() for c in cmd_list]

    def _get_file_completions(
        self, text: str, complete_event: CompleteEvent, only_files: bool = False
    ) -> Iterable[Completion]:
        # Logic:
        # - If text indicates path traversal (/, ., ~), use PathCompleter
        # - Else, check file count. If < 5000, use Fuzzy. Else use PathCompleter.

        if self._is_path_navigation(text):
            yield from self._get_path_completions(text, complete_event, only_files)
            return

        # Count files (cached strategy could be added here if needed)
        files = self._get_recursive_files(limit=5000)
        if len(files) < 5000:
            # Fuzzy Match
            yield from self._get_fuzzy_completions(text, files, only_files)
        else:
            # Fallback to PathCompleter for large repos
            yield from self._get_path_completions(text, complete_event, only_files)

    def _is_path_navigation(self, text: str) -> bool:
        return (
            text.startswith("/")
            or text.startswith(".")
            or text.startswith("~")
            or os.sep in text
        )

    def _get_path_completions(
        self, text: str, complete_event: CompleteEvent, only_files: bool
    ) -> Iterable[Completion]:
        # PathCompleter needs a document where text represents the path
        fake_document = Document(text=text, cursor_position=len(text))
        for c in self._path_completer.get_completions(fake_document, complete_event):
            if only_files:
                # Check if the completed path is a directory
                # Note: 'text' is the prefix. c.text is the completion suffix.
                # We need to reconstruct full path to check isdir
                # This is tricky with PathCompleter's internal logic.
                # A simple heuristic: if it ends with path separator, it's a dir.
                if c.text.endswith(os.sep):
                    continue
            yield c

    def _get_fuzzy_completions(
        self, text: str, files: list[str], only_files: bool
    ) -> Iterable[Completion]:
        matches = []
        for f in files:
            if only_files and f.endswith(os.sep):
                continue
            is_match, score = fuzzy_match(f, text)
            if is_match:
                matches.append((score, f))

        # Sort by score (lower is better)
        matches.sort(key=lambda x: x[0])

        # Return top 20
        for _, f in matches[:20]:
            yield Completion(f, start_position=-len(text))

    def _get_recursive_files(self, root: str = ".", limit: int = 5000) -> list[str]:
        # Simple walker with exclusions
        paths = []
        # Check if current dir is hidden
        cwd_is_hidden = os.path.basename(os.path.abspath(root)).startswith(".")

        try:
            for dirpath, dirnames, filenames in os.walk(root):
                # Exclude hidden directories unless root is hidden
                if not cwd_is_hidden:
                    dirnames[:] = [d for d in dirnames if not d.startswith(".")]

                # Exclude common ignores
                dirnames[:] = [
                    d
                    for d in dirnames
                    if d not in ("node_modules", "__pycache__", "venv", ".venv")
                ]

                rel_dir = os.path.relpath(dirpath, root)
                if rel_dir == ".":
                    rel_dir = ""

                # Add directories
                for d in dirnames:
                    paths.append(os.path.join(rel_dir, d) + os.sep)
                    if len(paths) >= limit:
                        return paths

                # Add files
                for f in filenames:
                    if not cwd_is_hidden and f.startswith("."):
                        continue
                    paths.append(os.path.join(rel_dir, f))
                    if len(paths) >= limit:
                        return paths
        except Exception:
            pass
        return paths



from zrb.llm.tool.bash import run_shell_command
from zrb.llm.tool.code import analyze_code
from zrb.llm.tool.file import (
    analyze_file,
    list_files,
    read_file,
    read_files,
    replace_in_file,
    search_files,
    write_file,
    write_files,
)
from zrb.llm.tool.mcp import load_mcp_config
from zrb.llm.tool.note import create_note_tools
from zrb.llm.tool.rag import create_rag_from_directory
from zrb.llm.tool.skill import create_activate_skill_tool
from zrb.llm.tool.sub_agent import create_sub_agent_tool
from zrb.llm.tool.web import open_web_page, search_internet
from zrb.llm.tool.zrb_task import create_list_zrb_task_tool, create_run_zrb_task_tool

__all__ = [
    "run_shell_command",
    "analyze_code",
    "list_files",
    "read_file",
    "read_files",
    "write_file",
    "write_files",
    "replace_in_file",
    "search_files",
    "analyze_file",
    "load_mcp_config",
    "create_note_tools",
    "create_rag_from_directory",
    "create_activate_skill_tool",
    "create_sub_agent_tool",
    "open_web_page",
    "search_internet",
    "create_list_zrb_task_tool",
    "create_run_zrb_task_tool",
]



from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from zrb.attr.type import BoolAttr, StrAttr, fstring
from zrb.config.config import CFG
from zrb.context.any_context import AnyContext
from zrb.context.print_fn import PrintFn
from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput
from zrb.llm.agent.agent import create_agent, run_agent
from zrb.llm.config.config import LLMConfig
from zrb.llm.config.config import llm_config as default_llm_config
from zrb.llm.config.limiter import LLMLimiter
from zrb.llm.config.limiter import llm_limiter as default_llm_limitter
from zrb.llm.history_manager.any_history_manager import AnyHistoryManager
from zrb.llm.history_manager.file_history_manager import FileHistoryManager
from zrb.llm.history_processor.summarizer import (
    summarize_history,
)
from zrb.llm.prompt.compose import PromptManager
from zrb.llm.util.attachment import get_attachments
from zrb.llm.util.stream_response import (
    create_event_handler,
    create_faint_printer,
)
from zrb.task.any_task import AnyTask
from zrb.task.base_task import BaseTask
from zrb.util.attr import get_attr, get_bool_attr
from zrb.util.string.name import get_random_name

if TYPE_CHECKING:
    from pydantic_ai import Tool, UserContent
    from pydantic_ai._agent_graph import HistoryProcessor
    from pydantic_ai.models import Model
    from pydantic_ai.settings import ModelSettings
    from pydantic_ai.tools import ToolFuncEither
    from pydantic_ai.toolsets import AbstractToolset


class LLMTask(BaseTask):

    def __init__(
        self,
        name: str,
        color: int | None = None,
        icon: str | None = None,
        description: str | None = None,
        cli_only: bool = False,
        input: list[AnyInput | None] | AnyInput | None = None,
        env: list[AnyEnv | None] | AnyEnv | None = None,
        system_prompt: (
            "Callable[[AnyContext], str | fstring | None] | str | None"
        ) = None,
        render_system_prompt: bool = False,
        prompt_manager: PromptManager | None = None,
        tools: list["Tool | ToolFuncEither"] = [],
        toolsets: list["AbstractToolset[None]"] = [],
        message: StrAttr | None = None,
        render_message: bool = True,
        attachment: "UserContent | list[UserContent] | Callable[[AnyContext], UserContent | list[UserContent]] | None" = None,  # noqa
        history_processors: list["HistoryProcessor"] = [],
        llm_config: LLMConfig | None = None,
        llm_limitter: LLMLimiter | None = None,
        model: (
            "Callable[[AnyContext], Model | str | fstring | None] | Model | None"
        ) = None,
        render_model: bool = True,
        model_settings: (
            "ModelSettings | Callable[[AnyContext], ModelSettings] | None"
        ) = None,
        conversation_name: StrAttr | None = None,
        render_conversation_name: bool = True,
        history_manager: AnyHistoryManager | None = None,
        tool_confirmation: Callable[[Any], Any] | None = None,
        yolo: BoolAttr = False,
        summarize_command: list[str] = [],
        execute_condition: bool | str | Callable[[AnyContext], bool] = True,
        retries: int = 2,
        retry_period: float = 0,
        readiness_check: list[AnyTask] | AnyTask | None = None,
        readiness_check_delay: float = 0.5,
        readiness_check_period: float = 5,
        readiness_failure_threshold: int = 1,
        readiness_timeout: int = 60,
        monitor_readiness: bool = False,
        upstream: list[AnyTask] | AnyTask | None = None,
        fallback: list[AnyTask] | AnyTask | None = None,
        successor: list[AnyTask] | AnyTask | None = None,
        print_fn: PrintFn | None = None,
    ):
        super().__init__(
            name=name,
            color=color,
            icon=icon,
            description=description,
            cli_only=cli_only,
            input=input,
            env=env,
            execute_condition=execute_condition,
            retries=retries,
            retry_period=retry_period,
            readiness_check=readiness_check,
            readiness_check_delay=readiness_check_delay,
            readiness_check_period=readiness_check_period,
            readiness_failure_threshold=readiness_failure_threshold,
            readiness_timeout=readiness_timeout,
            monitor_readiness=monitor_readiness,
            upstream=upstream,
            fallback=fallback,
            successor=successor,
            print_fn=print_fn,
        )
        self._llm_config = default_llm_config if llm_config is None else llm_config
        self._llm_limitter = (
            default_llm_limitter if llm_limitter is None else llm_limitter
        )
        self._system_prompt = system_prompt
        self._render_system_prompt = render_system_prompt
        self._prompt_manager = prompt_manager
        self._tools = tools
        self._toolsets = toolsets
        self._message = message
        self._render_message = render_message
        self._attachment = attachment
        self._history_processors = history_processors
        self._model = model
        self._render_model = render_model
        self._model_settings = model_settings
        self._conversation_name = conversation_name
        self._render_conversation_name = render_conversation_name
        self._history_manager = (
            FileHistoryManager(history_dir=CFG.LLM_HISTORY_DIR)
            if history_manager is None
            else history_manager
        )
        self._tool_confirmation = tool_confirmation
        self._yolo = yolo
        self._summarize_command = summarize_command

    @property
    def prompt_manager(self) -> PromptManager:
        if self._prompt_manager is None:
            raise ValueError(f"Task {self.name} doesn't have prompt_manager")
        return self._prompt_manager

    def add_toolset(self, *toolset: "AbstractToolset"):
        self.append_toolset(*toolset)

    def append_toolset(self, *toolset: "AbstractToolset"):
        self._toolsets += list(toolset)

    def add_tool(self, *tool: "Tool | ToolFuncEither"):
        self.append_tool(*tool)

    def append_tool(self, *tool: "Tool | ToolFuncEither"):
        self._tools += list(tool)

    def add_history_processor(self, *processor: "HistoryProcessor"):
        self.append_history_processor(*processor)

    def append_history_processor(self, *processor: "HistoryProcessor"):
        self._history_processors += list(processor)

    async def _exec_action(self, ctx: AnyContext) -> Any:
        conversation_name = self._get_conversation_name(ctx)
        message_history = self._history_manager.load(conversation_name)
        user_message = get_attr(ctx, self._message, "", self._render_message)
        user_attachments = get_attachments(ctx, self._attachment)

        if (
            isinstance(user_message, str)
            and user_message.strip() in self._summarize_command
        ):
            ctx.print("Compressing conversation history...", plain=True)
            new_history = await summarize_history(message_history)
            self._history_manager.update(conversation_name, new_history)
            self._history_manager.save(conversation_name)
            return "Conversation history compressed."

        yolo = get_bool_attr(ctx, self._yolo, False)
        system_prompt = self._get_system_prompt(ctx)
        ctx.log_debug(f"SYSTEM PROMPT: {system_prompt}")
        agent = create_agent(
            model=self._get_model(ctx),
            system_prompt=self._get_system_prompt(ctx),
            tools=self._tools,
            toolsets=self._toolsets,
            model_settings=self._get_model_settings(ctx),
            history_processors=self._history_processors,
            yolo=yolo,
        )

        print_event = create_faint_printer(ctx)
        handle_event = create_event_handler(
            print_event,
            show_tool_call_detail=CFG.LLM_SHOW_TOOL_CALL_DETAIL,
            show_tool_result=CFG.LLM_SHOW_TOOL_CALL_RESULT,
        )

        output, new_history = await run_agent(
            agent=agent,
            message=user_message,
            message_history=message_history,
            limiter=self._llm_limitter,
            attachments=user_attachments,
            print_fn=ctx.print,
            event_handler=handle_event,
            tool_confirmation=self._tool_confirmation,
        )

        self._history_manager.update(conversation_name, new_history)
        self._history_manager.save(conversation_name)
        ctx.log_debug(f"All messages: {new_history}")

        return output

    def _get_system_prompt(self, ctx: AnyContext) -> str:
        if self._prompt_manager is None:
            return str(
                get_attr(ctx, self._system_prompt, "", self._render_system_prompt)
            )
        compose_prompt = self._prompt_manager.compose_prompt()
        return compose_prompt(ctx)

    def _get_conversation_name(self, ctx: AnyContext) -> str:
        conversation_name = str(
            get_attr(ctx, self._conversation_name, "", self._render_conversation_name)
        )
        if conversation_name.strip() == "":
            conversation_name = get_random_name()
        return conversation_name

    def _get_model_settings(self, ctx: AnyContext) -> "ModelSettings | None":
        model_settings = self._model_settings
        rendered_model_settings = get_attr(ctx, model_settings, None)
        if rendered_model_settings is not None:
            return rendered_model_settings
        return self._llm_config.model_settings

    def _get_model(self, ctx: AnyContext) -> "str | Model":
        model = self._model
        rendered_model = get_attr(ctx, model, None, auto_render=self._render_model)
        if rendered_model is not None:
            return rendered_model
        return self._llm_config.model



from zrb.config.config import CFG


async def open_web_page(url: str) -> dict:
    """
    Downloads and converts a web page into clean, readable Markdown.

    **WHEN TO USE:**
    - To read specific articles, documentation, or blog posts.
    - To extract structured information from a known URL.

    **ARGS:**
    - `url`: The full web address to fetch.
    """
    html_content, links = await _fetch_page_content(url)
    markdown_content = _convert_html_to_markdown(html_content)
    return {"content": markdown_content, "links_on_page": links}


async def search_internet(
    query: str,
    page: int = 1,
) -> dict:
    """
    Performs a broad internet search.

    **WHEN TO USE:**
    - When you need to find information but don't have a specific URL.

    **ARGS:**
    - `query`: The search string or question.
    - `page`: Result page number (default 1).
    """
    if (
        CFG.SEARCH_INTERNET_METHOD.strip().lower() == "serpapi"
        and CFG.SERPAPI_KEY != ""
    ):
        from zrb.llm.tool.search.serpapi import search_internet as serpapi_search

        return serpapi_search(query, page=page)
    if (
        CFG.SEARCH_INTERNET_METHOD.strip().lower() == "brave"
        and CFG.BRAVE_API_KEY != ""
    ):
        from zrb.llm.tool.search.brave import search_internet as brave_search

        return brave_search(query, page=page)
    from zrb.llm.tool.search.searxng import search_internet as searxng_search

    return searxng_search(query, page=page)


async def _fetch_page_content(url: str) -> tuple:
    user_agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
    try:
        from playwright.async_api import async_playwright

        async with async_playwright() as p:
            browser = await p.chromium.launch(headless=True)
            page = await browser.new_page()
            await page.set_extra_http_headers({"User-Agent": user_agent})
            await page.goto(url, wait_until="networkidle", timeout=30000)
            content = await page.content()
            links = await page.eval_on_selector_all(
                "a[href]",
                "(elements, baseUrl) => elements.map(el => { const href = el.getAttribute('href'); if (!href || href.startsWith('#')) return null; try { return new URL(href, baseUrl).href; } catch (e) { return null; } }).filter(href => href !== null)",
                url,
            )
            await browser.close()
            return content, links
    except Exception:
        from urllib.parse import urljoin

        import requests
        from bs4 import BeautifulSoup

        response = requests.get(url, headers={"User-Agent": user_agent})
        response.raise_for_status()
        soup = BeautifulSoup(response.text, "html.parser")
        links = [
            urljoin(url, a["href"])
            for a in soup.find_all("a", href=True)
            if not a["href"].startswith("#")
        ]
        return response.text, links


def _convert_html_to_markdown(html_text: str) -> str:
    from bs4 import BeautifulSoup
    from markdownify import markdownify as md

    soup = BeautifulSoup(html_text, "html.parser")
    for tag in soup(
        ["script", "link", "meta", "style", "header", "footer", "nav", "aside"]
    ):
        tag.decompose()
    return md(str(soup))



from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from zrb.attr.type import BoolAttr, StrAttr, fstring
from zrb.config.config import CFG
from zrb.context.any_context import AnyContext
from zrb.context.print_fn import PrintFn
from zrb.context.shared_context import SharedContext
from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput
from zrb.input.bool_input import BoolInput
from zrb.input.str_input import StrInput
from zrb.llm.app.confirmation.handler import ConfirmationMiddleware
from zrb.llm.config.config import LLMConfig
from zrb.llm.config.config import llm_config as default_llm_config
from zrb.llm.config.limiter import LLMLimiter
from zrb.llm.history_manager.any_history_manager import AnyHistoryManager
from zrb.llm.history_manager.file_history_manager import FileHistoryManager
from zrb.llm.prompt.compose import PromptManager
from zrb.llm.task.llm_task import LLMTask
from zrb.llm.util.attachment import get_attachments
from zrb.session.session import Session
from zrb.task.any_task import AnyTask
from zrb.task.base_task import BaseTask
from zrb.util.attr import get_attr, get_bool_attr, get_str_attr
from zrb.util.string.name import get_random_name

if TYPE_CHECKING:
    from pydantic_ai import Tool, UserContent
    from pydantic_ai._agent_graph import HistoryProcessor
    from pydantic_ai.models import Model
    from pydantic_ai.settings import ModelSettings
    from pydantic_ai.tools import ToolFuncEither
    from pydantic_ai.toolsets import AbstractToolset
    from rich.theme import Theme


class LLMChatTask(BaseTask):

    def __init__(
        self,
        name: str,
        color: int | None = None,
        icon: str | None = None,
        description: str | None = None,
        cli_only: bool = False,
        input: list[AnyInput | None] | AnyInput | None = None,
        env: list[AnyEnv | None] | AnyEnv | None = None,
        system_prompt: (
            "Callable[[AnyContext], str | fstring | None] | str | None"
        ) = None,
        render_system_prompt: bool = False,
        prompt_manager: PromptManager | None = None,
        tools: list["Tool | ToolFuncEither"] = [],
        toolsets: list["AbstractToolset[None]"] = [],
        message: StrAttr | None = None,
        render_message: bool = True,
        attachment: "UserContent | list[UserContent] | Callable[[AnyContext], UserContent | list[UserContent]] | None" = None,  # noqa
        history_processors: list["HistoryProcessor"] = [],
        llm_config: LLMConfig | None = None,
        llm_limitter: LLMLimiter | None = None,
        model: (
            "Callable[[AnyContext], Model | str | fstring | None] | Model | None"
        ) = None,
        render_model: bool = True,
        model_settings: (
            "ModelSettings | Callable[[AnyContext], ModelSettings] | None"
        ) = None,
        conversation_name: StrAttr | None = None,
        render_conversation_name: bool = True,
        history_manager: AnyHistoryManager | None = None,
        tool_confirmation: Callable[[Any], Any] | None = None,
        yolo: BoolAttr = False,
        ui_summarize_commands: list[str] = [],
        ui_attach_commands: list[str] = [],
        ui_exit_commands: list[str] = [],
        ui_info_commands: list[str] = [],
        ui_save_commands: list[str] = [],
        ui_load_commands: list[str] = [],
        ui_redirect_output_commands: list[str] = [],
        ui_yolo_toggle_commands: list[str] = [],
        ui_exec_commands: list[str] = [],
        ui_greeting: StrAttr | None = None,
        render_ui_greeting: bool = True,
        ui_assistant_name: StrAttr | None = None,
        render_ui_assistant_name: bool = True,
        ui_jargon: StrAttr | None = None,
        render_ui_jargon: bool = True,
        ui_ascii_art: StrAttr | None = None,
        render_ui_ascii_art_name: bool = True,
        triggers: list[Callable[[], Any]] = [],
        confirmation_middlewares: list[ConfirmationMiddleware] = [],
        markdown_theme: "Theme | None" = None,
        interactive: BoolAttr = True,
        execute_condition: bool | str | Callable[[AnyContext], bool] = True,
        retries: int = 2,
        retry_period: float = 0,
        readiness_check: list[AnyTask] | AnyTask | None = None,
        readiness_check_delay: float = 0.5,
        readiness_check_period: float = 5,
        readiness_failure_threshold: int = 1,
        readiness_timeout: int = 60,
        monitor_readiness: bool = False,
        upstream: list[AnyTask] | AnyTask | None = None,
        fallback: list[AnyTask] | AnyTask | None = None,
        successor: list[AnyTask] | AnyTask | None = None,
        print_fn: PrintFn | None = None,
    ):
        super().__init__(
            name=name,
            color=color,
            icon=icon,
            description=description,
            cli_only=cli_only,
            input=input,
            env=env,
            execute_condition=execute_condition,
            retries=retries,
            retry_period=retry_period,
            readiness_check=readiness_check,
            readiness_check_delay=readiness_check_delay,
            readiness_check_period=readiness_check_period,
            readiness_failure_threshold=readiness_failure_threshold,
            readiness_timeout=readiness_timeout,
            monitor_readiness=monitor_readiness,
            upstream=upstream,
            fallback=fallback,
            successor=successor,
            print_fn=print_fn,
        )
        self._llm_config = default_llm_config if llm_config is None else llm_config
        self._llm_limitter = llm_limitter
        self._system_prompt = system_prompt
        self._render_system_prompt = render_system_prompt
        self._prompt_manager = prompt_manager
        self._tools = tools
        self._toolsets = toolsets
        self._message = message
        self._render_message = render_message
        self._attachment = attachment
        self._history_processors = history_processors
        self._model = model
        self._render_model = render_model
        self._model_settings = model_settings
        self._conversation_name = conversation_name
        self._render_conversation_name = render_conversation_name
        self._history_manager = (
            FileHistoryManager(history_dir=CFG.LLM_HISTORY_DIR)
            if history_manager is None
            else history_manager
        )
        self._tool_confirmation = tool_confirmation
        self._yolo = yolo
        self._ui_summarize_commands = ui_summarize_commands
        self._ui_attach_commands = ui_attach_commands
        self._ui_exit_commands = ui_exit_commands
        self._ui_info_commands = ui_info_commands
        self._ui_save_commands = ui_save_commands
        self._ui_load_commands = ui_load_commands
        self._ui_redirect_output_commands = ui_redirect_output_commands
        self._ui_yolo_toggle_commands = ui_yolo_toggle_commands
        self._ui_exec_commands = ui_exec_commands
        self._ui_greeting = ui_greeting
        self._render_ui_greeting = render_ui_greeting
        self._ui_assistant_name = ui_assistant_name
        self._render_ui_assistant_name = render_ui_assistant_name
        self._ui_jargon = ui_jargon
        self._render_ui_jargon = render_ui_jargon
        self._ui_ascii_art_name = ui_ascii_art
        self._render_ui_ascii_art_name = render_ui_ascii_art_name
        self._triggers = triggers
        self._confirmation_middlewares = confirmation_middlewares
        self._markdown_theme = markdown_theme
        self._interactive = interactive

    @property
    def prompt_manager(self) -> PromptManager:
        if self._prompt_manager is None:
            raise ValueError(f"Task {self.name} doesn't have prompt_manager")
        return self._prompt_manager

    def add_toolset(self, *toolset: "AbstractToolset"):
        self.append_toolset(*toolset)

    def append_toolset(self, *toolset: "AbstractToolset"):
        self._toolsets += list(toolset)

    def add_tool(self, *tool: "Tool | ToolFuncEither"):
        self.append_tool(*tool)

    def append_tool(self, *tool: "Tool | ToolFuncEither"):
        self._tools += list(tool)

    def add_history_processor(self, *processor: "HistoryProcessor"):
        self.append_history_processor(*processor)

    def append_history_processor(self, *processor: "HistoryProcessor"):
        self._history_processors += list(processor)

    def add_confirmation_middleware(self, *middleware: ConfirmationMiddleware):
        self.prepend_confirmation_middleware(*middleware)

    def prepend_confirmation_middleware(self, *middleware: ConfirmationMiddleware):
        self._confirmation_middlewares = (
            list(middleware) + self._confirmation_middlewares
        )

    def add_trigger(self, *trigger: Callable[[], Any]):
        self.append_trigger(*trigger)

    def append_trigger(self, *trigger: Callable[[], Any]):
        self._triggers += trigger

    async def _exec_action(self, ctx: AnyContext) -> Any:
        from zrb.llm.app.lexer import CLIStyleLexer
        from zrb.llm.app.ui import UI

        initial_conversation_name = self._get_conversation_name(ctx)
        initial_yolo = get_bool_attr(ctx, self._yolo, False)
        initial_message = get_attr(ctx, self._message, "", self._render_message)
        initial_attachments = get_attachments(ctx, self._attachment)
        ui_greeting = get_str_attr(ctx, self._ui_greeting, "", self._render_ui_greeting)
        ui_assistant_name = get_str_attr(
            ctx, self._ui_assistant_name, "", self._render_ui_assistant_name
        )
        ui_jargon = get_str_attr(ctx, self._ui_jargon, "", self._render_ui_jargon)
        ascii_art = get_str_attr(
            ctx, self._ui_ascii_art_name, "", self._render_ui_ascii_art_name
        )
        interactive = get_bool_attr(ctx, self._interactive, True)

        llm_task_core = LLMTask(
            name=f"{self.name}-process",
            input=[
                StrInput("message", "Message"),
                StrInput("session", "Conversation Session"),
                BoolInput("yolo", "YOLO Mode"),
                StrInput("attachments", "Attachments"),
            ],
            env=self.envs,
            system_prompt=self._system_prompt,
            render_system_prompt=self._render_system_prompt,
            prompt_manager=self._prompt_manager,
            tools=self._tools,
            toolsets=self._toolsets,
            history_processors=self._history_processors,
            llm_config=self._llm_config,
            llm_limitter=self._llm_limitter,
            model=self._model,
            render_model=self._render_model,
            model_settings=self._model_settings,
            history_manager=self._history_manager,
            message="{ctx.input.message}",
            conversation_name="{ctx.input.session}",
            yolo="{ctx.input.yolo}",
            attachment=lambda ctx: ctx.input.attachments,
            summarize_command=self._ui_summarize_commands,
        )

        if not interactive:
            session_input = {
                "message": initial_message,
                "session": initial_conversation_name,
                "yolo": initial_yolo,
                "attachments": initial_attachments,
            }
            shared_ctx = SharedContext(
                input=session_input,
                print_fn=ctx.shared_print,  # Use current task's print function
            )
            session = Session(shared_ctx)
            return await llm_task_core.async_run(session)

        ui = UI(
            greeting=ui_greeting,
            assistant_name=ui_assistant_name,
            ascii_art=ascii_art,
            jargon=ui_jargon,
            output_lexer=CLIStyleLexer(),
            llm_task=llm_task_core,
            history_manager=self._history_manager,
            initial_message=initial_message,
            initial_attachments=initial_attachments,
            conversation_session_name=initial_conversation_name,
            yolo=initial_yolo,
            triggers=self._triggers,
            confirmation_middlewares=self._confirmation_middlewares,
            markdown_theme=self._markdown_theme,
            summarize_commands=self._ui_summarize_commands,
            attach_commands=self._ui_attach_commands,
            exit_commands=self._ui_exit_commands,
            info_commands=self._ui_info_commands,
            save_commands=self._ui_save_commands,
            load_commands=self._ui_load_commands,
            yolo_toggle_commands=self._ui_yolo_toggle_commands,
            redirect_output_commands=self._ui_redirect_output_commands,
            exec_commands=self._ui_exec_commands,
            model=self._get_model(ctx),
        )
        await ui.run_async()
        return ui.last_output

    def _get_conversation_name(self, ctx: AnyContext) -> str:
        conversation_name = str(
            get_attr(ctx, self._conversation_name, "", self._render_conversation_name)
        )
        if conversation_name.strip() == "":
            conversation_name = get_random_name()
        return conversation_name

    def _get_model(self, ctx: AnyContext) -> "str | Model":
        model = self._model
        rendered_model = get_attr(ctx, model, None, auto_render=self._render_model)
        if rendered_model is not None:
            return rendered_model
        return self._llm_config.model



from zrb.llm.agent.agent import create_agent, run_agent, tool_confirmation_var
from zrb.llm.agent.summarizer import create_summarizer_agent

__all__ = [
    "create_agent",
    "run_agent",
    "tool_confirmation_var",
    "create_summarizer_agent",
]



import os
from typing import Dict, List, Optional

IGNORE_DIRS = {
    ".git",
    "node_modules",
    "__pycache__",
    ".venv",
    "venv",
    "dist",
    "build",
    ".idea",
    ".vscode",
}


class Skill:
    def __init__(self, name: str, path: str, description: str):
        self.name = name
        self.path = path
        self.description = description


class SkillManager:
    def __init__(self, root_dir: str = "."):
        self.root_dir = root_dir
        self._skills: Dict[str, Skill] = {}

    def scan(self) -> List[Skill]:
        self._skills = {}
        for root, dirs, files in os.walk(self.root_dir):
            dirs[:] = [d for d in dirs if d not in IGNORE_DIRS]

            for file in files:
                if file == "SKILL.md" or file.endswith(".skill.md"):
                    full_path = os.path.join(root, file)
                    rel_path = os.path.relpath(full_path, self.root_dir)
                    self._load_skill(rel_path, full_path)

        return list(self._skills.values())

    def _load_skill(self, rel_path: str, full_path: str):
        try:
            with open(full_path, "r", encoding="utf-8") as f:
                lines = f.readlines()

            name = os.path.basename(os.path.dirname(full_path))
            description = "No description"

            # Parse Markdown for Header 1
            for line in lines:
                stripped = line.strip()
                if stripped.startswith("# "):
                    name = stripped[2:].strip()
                    break

            # Use name as key, handle duplicates
            key = name
            if key in self._skills:
                key = f"{name} ({rel_path})"

            self._skills[key] = Skill(name=key, path=rel_path, description=description)

        except Exception:
            pass

    def get_skill_content(self, name: str) -> Optional[str]:
        skill = self._skills.get(name)
        if not skill:
            # Try partial match or path match
            for s in self._skills.values():
                if s.name == name or s.path == name:
                    skill = s
                    break

        if not skill:
            return None

        try:
            with open(skill.path, "r", encoding="utf-8") as f:
                return f.read()
        except Exception as e:
            return f"Error reading skill file: {e}"


skill_manager = SkillManager()



import asyncio
import json
import time
from collections import deque
from typing import Any, Callable

from zrb.config.config import CFG


class LLMLimiter:
    """
    Manages LLM constraints: Context Window (Pruning) and Rate Limits (Throttling).
    Designed as a singleton to share limits across tasks.
    """

    def __init__(self):
        # Sliding window logs
        self._request_log: deque[float] = deque()
        self._token_log: deque[tuple[float, int]] = deque()

        # Internal overrides
        self._max_requests_per_minute: int | None = None
        self._max_tokens_per_minute: int | None = None
        self._max_tokens_per_request: int | None = None
        self._throttle_check_interval: float | None = None

    # --- Configuration Properties ---

    @property
    def max_requests_per_minute(self) -> int:
        if self._max_requests_per_minute is not None:
            return self._max_requests_per_minute
        return getattr(CFG, "LLM_MAX_REQUESTS_PER_MINUTE", None) or 60

    @max_requests_per_minute.setter
    def max_requests_per_minute(self, value: int):
        self._max_requests_per_minute = value

    @property
    def max_tokens_per_minute(self) -> int:
        if self._max_tokens_per_minute is not None:
            return self._max_tokens_per_minute
        return getattr(CFG, "LLM_MAX_TOKENS_PER_MINUTE", None) or 100_000

    @max_tokens_per_minute.setter
    def max_tokens_per_minute(self, value: int):
        self._max_tokens_per_minute = value

    @property
    def max_tokens_per_request(self) -> int:
        if self._max_tokens_per_request is not None:
            return self._max_tokens_per_request
        return getattr(CFG, "LLM_MAX_TOKENS_PER_REQUEST", None) or 16_000

    @max_tokens_per_request.setter
    def max_tokens_per_request(self, value: int):
        self._max_tokens_per_request = value

    @property
    def throttle_check_interval(self) -> float:
        if self._throttle_check_interval is not None:
            return self._throttle_check_interval
        return getattr(CFG, "LLM_THROTTLE_SLEEP", None) or 0.1

    @throttle_check_interval.setter
    def throttle_check_interval(self, value: float):
        self._throttle_check_interval = value

    @property
    def use_tiktoken(self) -> bool:
        return getattr(CFG, "USE_TIKTOKEN", False)

    @property
    def tiktoken_encoding(self) -> str:
        return getattr(CFG, "TIKTOKEN_ENCODING_NAME", "cl100k_base")

    # --- Public API ---

    def fit_context_window(self, history: list[Any], new_message: Any) -> list[Any]:
        """
        Prunes the history (removing oldest turns) so that 'history + new_message'
        fits within 'max_tokens_per_request'.
        Ensures strict tool call pairing by removing full conversation turns.
        """
        if not history:
            return history

        # Import message types locally to avoid circular deps or startup cost
        try:
            from pydantic_ai.messages import (
                ModelRequest,
                ToolReturnPart,
                UserPromptPart,
            )
        except ImportError:
            # Fallback if pydantic_ai is not installed (unlikely in context)
            return []

        def is_turn_start(msg: Any) -> bool:
            """Identify start of a new user interaction (User Prompt without Tool Return)."""
            if not isinstance(msg, ModelRequest):
                return False
            has_user = any(isinstance(p, UserPromptPart) for p in msg.parts)
            has_return = any(isinstance(p, ToolReturnPart) for p in msg.parts)
            return has_user and not has_return

        new_msg_tokens = self._count_tokens(new_message)
        if new_msg_tokens > self.max_tokens_per_request:
            return []

        pruned_history = list(history)

        while pruned_history:
            history_tokens = self._count_tokens(pruned_history)
            total_tokens = history_tokens + new_msg_tokens

            if total_tokens <= self.max_tokens_per_request:
                break

            # Pruning Strategy: Find the start of the *next* turn and cut everything before it.
            # We start searching from index 1 because removing index 0 (current start) is the goal.
            next_turn_index = -1
            for i in range(1, len(pruned_history)):
                if is_turn_start(pruned_history[i]):
                    next_turn_index = i
                    break

            if next_turn_index != -1:
                # Remove everything up to the next turn
                pruned_history = pruned_history[next_turn_index:]
            else:
                # No subsequent turns found.
                # This implies the history contains only one (potentially long) turn or partial fragments.
                # To satisfy the limit, we must clear the history entirely.
                pruned_history = []

        return pruned_history

    async def acquire(self, content: Any, notifier: Callable[[str], Any] | None = None):
        """
        Acquires permission to proceed with the given content.
        Calculates token count internally and waits if rate limits are exceeded.
        """
        # Calculate tokens once
        estimated_tokens = self._count_tokens(content)

        # 1. Prune logs older than 60 seconds
        self._prune_logs()

        # 2. Check limits loop
        notified = False
        while not self._can_proceed(estimated_tokens):
            wait_time = self._calculate_wait_time(estimated_tokens)
            reason = self._get_limit_reason(estimated_tokens)

            if notifier:
                msg = f"Rate Limit Reached: {reason}. Waiting {wait_time:.1f}s..."
                # Only notify once or if status changes? Simple is better.
                notifier(msg)
                notified = True

            await asyncio.sleep(self.throttle_check_interval)
            self._prune_logs()

        if notified and notifier:
            notifier("")  # Clear status

        # 3. Record usage
        now = time.time()
        self._request_log.append(now)
        self._token_log.append((now, estimated_tokens))

    def count_tokens(self, content: Any) -> int:
        """Public alias for internal counter."""
        return self._count_tokens(content)

    # --- Internal Helpers ---

    def _count_tokens(self, content: Any) -> int:
        text = self._to_str(content)
        if self.use_tiktoken:
            try:
                import tiktoken

                enc = tiktoken.get_encoding(self.tiktoken_encoding)
                return len(enc.encode(text))
            except ImportError:
                pass
        # Fallback approximation (char/4)
        return len(text) // 4

    def _to_str(self, content: Any) -> str:
        if isinstance(content, str):
            return content
        try:
            return json.dumps(content, default=str)
        except Exception:
            return str(content)

    def _prune_logs(self):
        now = time.time()
        window_start = now - 60

        while self._request_log and self._request_log[0] < window_start:
            self._request_log.popleft()

        while self._token_log and self._token_log[0][0] < window_start:
            self._token_log.popleft()

    def _can_proceed(self, tokens: int) -> bool:
        requests_ok = len(self._request_log) < self.max_requests_per_minute

        current_tokens = sum(t for _, t in self._token_log)
        tokens_ok = (current_tokens + tokens) <= self.max_tokens_per_minute

        return requests_ok and tokens_ok

    def _get_limit_reason(self, tokens: int) -> str:
        if len(self._request_log) >= self.max_requests_per_minute:
            return f"Max Requests ({self.max_requests_per_minute}/min)"
        return f"Max Tokens ({self.max_tokens_per_minute}/min)"

    def _calculate_wait_time(self, tokens: int) -> float:
        now = time.time()
        # Default wait
        wait = 1.0

        # If request limit hit, wait until oldest request expires
        if len(self._request_log) >= self.max_requests_per_minute:
            oldest = self._request_log[0]
            wait = max(0.1, 60 - (now - oldest))

        # If token limit hit, wait until enough tokens expire
        current_tokens = sum(t for _, t in self._token_log)
        if current_tokens + tokens > self.max_tokens_per_minute:
            needed = (current_tokens + tokens) - self.max_tokens_per_minute
            freed = 0
            for ts, count in self._token_log:
                freed += count
                if freed >= needed:
                    wait = max(wait, 60 - (now - ts))
                    break

        return wait


llm_limiter = LLMLimiter()



import importlib.metadata as metadata
import logging
import os

from zrb.config.helper import (
    get_current_shell,
    get_default_diff_edit_command,
    get_env,
    get_log_level,
    get_max_token_threshold,
    limit_token_threshold,
)
from zrb.util.string.conversion import to_boolean
from zrb.util.string.format import fstring_format

_DEFAULT_BANNER = """
                bb
   zzzzz rr rr  bb
     zz  rrr  r bbbbbb
    zz   rr     bb   bb
   zzzzz rr     bbbbbb   {VERSION} Pollux
   _ _ . .  . _ .  _ . . .
Your Automation Powerhouse
☕ Donate at: https://stalchmst.com
🐙 Submit issues/PR at: https://github.com/state-alchemists/zrb
🐤 Follow us at: https://twitter.com/zarubastalchmst
"""


class Config:
    def __init__(self):
        self.DEFAULT_ENV_PREFIX = "ZRB"
        self.DEFAULT_SHELL = ""
        self.DEFAULT_EDITOR = "nano"
        self.DEFAULT_EDIT_COMMAND_TPL = ""
        self.DEFAULT_INIT_MODULES = ""
        self.DEFAULT_ROOT_GROUP_NAME = "zrb"
        self.DEFAULT_ROOT_GROUP_DESCRIPTION = "Your Automation Powerhouse"
        self.DEFAULT_INIT_SCRIPTS = ""
        self.DEFAULT_INIT_FILE_NAME = "zrb_init.py"
        self.DEFAULT_LOGGING_LEVEL = "WARNING"
        self.DEFAULT_LOAD_BUILTIN = "1"
        self.DEFAULT_WARN_UNRECOMMENDED_COMMAND = "1"
        self.DEFAULT_SESSION_LOG_DIR = ""
        self.DEFAULT_TODO_DIR = ""
        self.DEFAULT_TODO_VISUAL_FILTER = ""
        self.DEFAULT_TODO_RETENTION = "2w"
        self.DEFAULT_VERSION = ""
        self.DEFAULT_WEB_CSS_PATH = ""
        self.DEFAULT_WEB_JS_PATH = ""
        self.DEFAULT_WEB_FAVICON_PATH = "/static/favicon-32x32.png"
        self.DEFAULT_WEB_COLOR = ""
        self.DEFAULT_WEB_HTTP_PORT = "21213"
        self.DEFAULT_WEB_GUEST_USERNAME = "user"
        self.DEFAULT_WEB_SUPER_ADMIN_USERNAME = "admin"
        self.DEFAULT_WEB_SUPER_ADMIN_PASSWORD = "admin"
        self.DEFAULT_WEB_ACCESS_TOKEN_COOKIE_NAME = "access_token"
        self.DEFAULT_WEB_REFRESH_TOKEN_COOKIE_NAME = "refresh_token"
        self.DEFAULT_WEB_SECRET_KEY = "zrb"
        self.DEFAULT_WEB_ENABLE_AUTH = "0"
        self.DEFAULT_WEB_AUTH_ACCESS_TOKEN_EXPIRE_MINUTES = "30"
        self.DEFAULT_WEB_AUTH_REFRESH_TOKEN_EXPIRE_MINUTES = "60"
        self.DEFAULT_WEB_TITLE = "Zrb"
        self.DEFAULT_WEB_JARGON = "Your Automation PowerHouse"
        self.DEFAULT_WEB_HOMEPAGE_INTRO = "Welcome to Zrb Web Interface"
        self.DEFAULT_LLM_ASSISTANT_NAME = ""
        self.DEFAULT_LLM_ASSISTANT_ASCII_ART = "default"
        self.DEFAULT_LLM_ASSISTANT_JARGON = ""
        self.DEFAULT_LLM_HISTORY_DIR = ""
        self.DEFAULT_LLM_NOTE_FILE = ""
        self.DEFAULT_LLM_MODEL = ""
        self.DEFAULT_LLM_BASE_URL = ""
        self.DEFAULT_LLM_API_KEY = ""
        self.DEFAULT_LLM_MAX_REQUESTS_PER_MINUTE = "60"
        self.DEFAULT_LLM_MAX_TOKENS_PER_MINUTE = "120000"
        self.DEFAULT_LLM_MAX_TOKENS_PER_REQUEST = "120000"
        self.DEFAULT_LLM_THROTTLE_SLEEP = "1.0"
        self.DEFAULT_LLM_HISTORY_SUMMARIZATION_WINDOW = "5"
        self.DEFAULT_LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD = ""
        self.DEFAULT_LLM_REPO_ANALYSIS_EXTRACTION_TOKEN_THRESHOLD = ""
        self.DEFAULT_LLM_REPO_ANALYSIS_SUMMARIZATION_TOKEN_THRESHOLD = ""
        self.DEFAULT_LLM_FILE_ANALYSIS_TOKEN_THRESHOLD = ""
        self.DEFAULT_LLM_PROMPT_DIR = ""
        self.DEFAULT_RAG_EMBEDDING_API_KEY = ""
        self.DEFAULT_RAG_EMBEDDING_BASE_URL = ""
        self.DEFAULT_RAG_EMBEDDING_MODEL = "text-embedding-ada-002"
        self.DEFAULT_RAG_CHUNK_SIZE = "1024"
        self.DEFAULT_RAG_OVERLAP = "128"
        self.DEFAULT_RAG_MAX_RESULT_COUNT = "5"
        self.DEFAULT_SEARCH_INTERNET_METHOD = "serpapi"
        self.DEFAULT_BRAVE_API_KEY = ""
        self.DEFAULT_BRAVE_API_SAFE = "off"
        self.DEFAULT_BRAVE_API_LANG = "en"
        self.DEFAULT_SERPAPI_KEY = ""
        self.DEFAULT_SERPAPI_SAFE = "off"
        self.DEFAULT_SERPAPI_LANG = "en"
        self.DEFAULT_SEARXNG_PORT = "8080"
        self.DEFAULT_SEARXNG_BASE_URL = ""
        self.DEFAULT_SEARXNG_SAFE = "0"
        self.DEFAULT_SEARXNG_LANG = "en"
        self.DEFAULT_BANNER = _DEFAULT_BANNER
        self.DEFAULT_LLM_SHOW_TOOL_CALL_DETAIL = "0"
        self.DEFAULT_LLM_SHOW_TOOL_CALL_RESULT = "0"
        self.DEFAULT_USE_TIKTOKEN = "1"
        self.DEFAULT_TIKTOKEN_ENCODING_NAME = "cl100k_base"
        self.DEFAULT_MCP_CONFIG_FILE = "mcp.json"

    @property
    def ENV_PREFIX(self) -> str:
        return os.getenv("_ZRB_ENV_PREFIX", self.DEFAULT_ENV_PREFIX)

    @ENV_PREFIX.setter
    def ENV_PREFIX(self, value: str):
        os.environ["_ZRB_ENV_PREFIX"] = value

    @property
    def LOGGER(self) -> logging.Logger:
        return logging.getLogger()

    @property
    def SHELL(self) -> str:
        return get_env(
            "SHELL",
            (self.DEFAULT_SHELL if self.DEFAULT_SHELL != "" else get_current_shell()),
            self.ENV_PREFIX,
        )

    @SHELL.setter
    def SHELL(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_SHELL"] = value

    @property
    def EDITOR(self) -> str:
        return get_env("EDITOR", self.DEFAULT_EDITOR, self.ENV_PREFIX)

    @EDITOR.setter
    def EDITOR(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_EDITOR"] = value

    @property
    def DIFF_EDIT_COMMAND_TPL(self) -> str:
        return get_env(
            "DIFF_EDIT_COMMAND",
            (
                self.DEFAULT_EDIT_COMMAND_TPL
                if self.DEFAULT_EDIT_COMMAND_TPL != ""
                else get_default_diff_edit_command(self.EDITOR)
            ),
            self.ENV_PREFIX,
        )

    @DIFF_EDIT_COMMAND_TPL.setter
    def DIFF_EDIT_COMMAND_TPL(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_DIFF_EDIT_COMMAND"] = value

    @property
    def INIT_MODULES(self) -> list[str]:
        init_modules_str = get_env(
            "INIT_MODULES", self.DEFAULT_INIT_MODULES, self.ENV_PREFIX
        )
        if init_modules_str != "":
            return [
                module.strip()
                for module in init_modules_str.split(":")
                if module.strip() != ""
            ]
        return []

    @INIT_MODULES.setter
    def INIT_MODULES(self, value: list[str]):
        os.environ[f"{self.ENV_PREFIX}_INIT_MODULES"] = ":".join(value)

    @property
    def ROOT_GROUP_NAME(self) -> str:
        return get_env("ROOT_GROUP_NAME", self.DEFAULT_ROOT_GROUP_NAME, self.ENV_PREFIX)

    @ROOT_GROUP_NAME.setter
    def ROOT_GROUP_NAME(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_ROOT_GROUP_NAME"] = value

    @property
    def ROOT_GROUP_DESCRIPTION(self) -> str:
        return get_env(
            "ROOT_GROUP_DESCRIPTION",
            self.DEFAULT_ROOT_GROUP_DESCRIPTION,
            self.ENV_PREFIX,
        )

    @ROOT_GROUP_DESCRIPTION.setter
    def ROOT_GROUP_DESCRIPTION(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_ROOT_GROUP_DESCRIPTION"] = value

    @property
    def INIT_SCRIPTS(self) -> list[str]:
        init_scripts_str = get_env(
            "INIT_SCRIPTS", self.DEFAULT_INIT_SCRIPTS, self.ENV_PREFIX
        )
        if init_scripts_str != "":
            return [
                script.strip()
                for script in init_scripts_str.split(":")
                if script.strip() != ""
            ]
        return []

    @INIT_SCRIPTS.setter
    def INIT_SCRIPTS(self, value: list[str]):
        os.environ[f"{self.ENV_PREFIX}_INIT_SCRIPTS"] = ":".join(value)

    @property
    def INIT_FILE_NAME(self) -> str:
        return get_env("INIT_FILE_NAME", self.DEFAULT_INIT_FILE_NAME, self.ENV_PREFIX)

    @INIT_FILE_NAME.setter
    def INIT_FILE_NAME(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_INIT_FILE_NAME"] = value

    @property
    def LOGGING_LEVEL(self) -> int:
        return get_log_level(
            get_env("LOGGING_LEVEL", self.DEFAULT_LOGGING_LEVEL, self.ENV_PREFIX)
        )

    @LOGGING_LEVEL.setter
    def LOGGING_LEVEL(self, value: int | str):
        if isinstance(value, int):
            value = logging.getLevelName(value)
        os.environ[f"{self.ENV_PREFIX}_LOGGING_LEVEL"] = str(value)

    @property
    def LOAD_BUILTIN(self) -> bool:
        return to_boolean(
            get_env("LOAD_BUILTIN", self.DEFAULT_LOAD_BUILTIN, self.ENV_PREFIX)
        )

    @LOAD_BUILTIN.setter
    def LOAD_BUILTIN(self, value: bool):
        os.environ[f"{self.ENV_PREFIX}_LOAD_BUILTIN"] = "1" if value else "0"

    @property
    def WARN_UNRECOMMENDED_COMMAND(self) -> bool:
        return to_boolean(
            get_env(
                "WARN_UNRECOMMENDED_COMMAND",
                self.DEFAULT_WARN_UNRECOMMENDED_COMMAND,
                self.ENV_PREFIX,
            )
        )

    @WARN_UNRECOMMENDED_COMMAND.setter
    def WARN_UNRECOMMENDED_COMMAND(self, value: bool):
        os.environ[f"{self.ENV_PREFIX}_WARN_UNRECOMMENDED_COMMAND"] = (
            "1" if value else "0"
        )

    @property
    def SESSION_LOG_DIR(self) -> str:
        default = self.DEFAULT_SESSION_LOG_DIR
        if default == "":
            default = os.path.expanduser(
                os.path.join("~", f".{self.ROOT_GROUP_NAME}-session")
            )
        return get_env(
            "SESSION_LOG_DIR",
            default,
            self.ENV_PREFIX,
        )

    @SESSION_LOG_DIR.setter
    def SESSION_LOG_DIR(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_SESSION_LOG_DIR"] = value

    @property
    def TODO_DIR(self) -> str:
        default = self.DEFAULT_TODO_DIR
        if default == "":
            default = os.path.expanduser(os.path.join("~", "todo"))
        return get_env(
            "TODO_DIR",
            default,
            self.ENV_PREFIX,
        )

    @TODO_DIR.setter
    def TODO_DIR(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_TODO_DIR"] = value

    @property
    def TODO_VISUAL_FILTER(self) -> str:
        return get_env("TODO_FILTER", self.DEFAULT_TODO_VISUAL_FILTER, self.ENV_PREFIX)

    @TODO_VISUAL_FILTER.setter
    def TODO_VISUAL_FILTER(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_TODO_FILTER"] = value

    @property
    def TODO_RETENTION(self) -> str:
        return get_env("TODO_RETENTION", self.DEFAULT_TODO_RETENTION, self.ENV_PREFIX)

    @TODO_RETENTION.setter
    def TODO_RETENTION(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_TODO_RETENTION"] = value

    @property
    def VERSION(self) -> str:
        custom_version = os.getenv("_ZRB_CUSTOM_VERSION", "")
        if custom_version != "":
            return custom_version
        return (
            self.DEFAULT_VERSION
            if self.DEFAULT_VERSION != ""
            else metadata.version("zrb")
        )

    @VERSION.setter
    def VERSION(self, value: str):
        os.environ["_ZRB_CUSTOM_VERSION"] = value

    @property
    def WEB_CSS_PATH(self) -> list[str]:
        web_css_path_str = get_env(
            "WEB_CSS_PATH", self.DEFAULT_WEB_CSS_PATH, self.ENV_PREFIX
        )
        if web_css_path_str != "":
            return [
                path.strip()
                for path in web_css_path_str.split(":")
                if path.strip() != ""
            ]
        return []

    @WEB_CSS_PATH.setter
    def WEB_CSS_PATH(self, value: list[str]):
        os.environ[f"{self.ENV_PREFIX}_WEB_CSS_PATH"] = ":".join(value)

    @property
    def WEB_JS_PATH(self) -> list[str]:
        web_js_path_str = get_env(
            "WEB_JS_PATH", self.DEFAULT_WEB_JS_PATH, self.ENV_PREFIX
        )
        if web_js_path_str != "":
            return [
                path.strip()
                for path in web_js_path_str.split(":")
                if path.strip() != ""
            ]
        return []

    @WEB_JS_PATH.setter
    def WEB_JS_PATH(self, value: list[str]):
        os.environ[f"{self.ENV_PREFIX}_WEB_JS_PATH"] = ":".join(value)

    @property
    def WEB_FAVICON_PATH(self) -> str:
        return get_env(
            "WEB_FAVICON_PATH", self.DEFAULT_WEB_FAVICON_PATH, self.ENV_PREFIX
        )

    @WEB_FAVICON_PATH.setter
    def WEB_FAVICON_PATH(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_WEB_FAVICON_PATH"] = value

    @property
    def WEB_COLOR(self) -> str:
        return get_env("WEB_COLOR", self.DEFAULT_WEB_COLOR, self.ENV_PREFIX)

    @WEB_COLOR.setter
    def WEB_COLOR(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_WEB_COLOR"] = value

    @property
    def WEB_HTTP_PORT(self) -> int:
        return int(
            get_env("WEB_HTTP_PORT", self.DEFAULT_WEB_HTTP_PORT, self.ENV_PREFIX)
        )

    @WEB_HTTP_PORT.setter
    def WEB_HTTP_PORT(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_WEB_HTTP_PORT"] = str(value)

    @property
    def WEB_GUEST_USERNAME(self) -> str:
        return get_env(
            "WEB_GUEST_USERNAME", self.DEFAULT_WEB_GUEST_USERNAME, self.ENV_PREFIX
        )

    @WEB_GUEST_USERNAME.setter
    def WEB_GUEST_USERNAME(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_WEB_GUEST_USERNAME"] = value

    @property
    def WEB_SUPER_ADMIN_USERNAME(self) -> str:
        return get_env(
            "WEB_SUPER_ADMIN_USERNAME",
            self.DEFAULT_WEB_SUPER_ADMIN_USERNAME,
            self.ENV_PREFIX,
        )

    @WEB_SUPER_ADMIN_USERNAME.setter
    def WEB_SUPER_ADMIN_USERNAME(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_WEB_SUPER_ADMIN_USERNAME"] = value

    @property
    def WEB_SUPER_ADMIN_PASSWORD(self) -> str:
        return get_env(
            "WEB_SUPER_ADMIN_PASSWORD",
            self.DEFAULT_WEB_SUPER_ADMIN_PASSWORD,
            self.ENV_PREFIX,
        )

    @WEB_SUPER_ADMIN_PASSWORD.setter
    def WEB_SUPER_ADMIN_PASSWORD(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_WEB_SUPER_ADMIN_PASSWORD"] = value

    @property
    def WEB_ACCESS_TOKEN_COOKIE_NAME(self) -> str:
        return get_env(
            "WEB_ACCESS_TOKEN_COOKIE_NAME",
            self.DEFAULT_WEB_ACCESS_TOKEN_COOKIE_NAME,
            self.ENV_PREFIX,
        )

    @WEB_ACCESS_TOKEN_COOKIE_NAME.setter
    def WEB_ACCESS_TOKEN_COOKIE_NAME(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_WEB_ACCESS_TOKEN_COOKIE_NAME"] = value

    @property
    def WEB_REFRESH_TOKEN_COOKIE_NAME(self) -> str:
        return get_env(
            "WEB_REFRESH_TOKEN_COOKIE_NAME",
            self.DEFAULT_WEB_REFRESH_TOKEN_COOKIE_NAME,
            self.ENV_PREFIX,
        )

    @WEB_REFRESH_TOKEN_COOKIE_NAME.setter
    def WEB_REFRESH_TOKEN_COOKIE_NAME(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_WEB_REFRESH_TOKEN_COOKIE_NAME"] = value

    @property
    def WEB_SECRET_KEY(self) -> str:
        return get_env("WEB_SECRET", self.DEFAULT_WEB_SECRET_KEY, self.ENV_PREFIX)

    @WEB_SECRET_KEY.setter
    def WEB_SECRET_KEY(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_WEB_SECRET"] = value

    @property
    def WEB_ENABLE_AUTH(self) -> bool:
        return to_boolean(
            get_env("WEB_ENABLE_AUTH", self.DEFAULT_WEB_ENABLE_AUTH, self.ENV_PREFIX)
        )

    @WEB_ENABLE_AUTH.setter
    def WEB_ENABLE_AUTH(self, value: bool):
        os.environ[f"{self.ENV_PREFIX}_WEB_ENABLE_AUTH"] = "1" if value else "0"

    @property
    def WEB_AUTH_ACCESS_TOKEN_EXPIRE_MINUTES(self) -> int:
        return int(
            get_env(
                "WEB_ACCESS_TOKEN_EXPIRE_MINUTES",
                self.DEFAULT_WEB_AUTH_ACCESS_TOKEN_EXPIRE_MINUTES,
                self.ENV_PREFIX,
            )
        )

    @WEB_AUTH_ACCESS_TOKEN_EXPIRE_MINUTES.setter
    def WEB_AUTH_ACCESS_TOKEN_EXPIRE_MINUTES(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_WEB_ACCESS_TOKEN_EXPIRE_MINUTES"] = str(value)

    @property
    def WEB_AUTH_REFRESH_TOKEN_EXPIRE_MINUTES(self) -> int:
        return int(
            get_env(
                "WEB_REFRESH_TOKEN_EXPIRE_MINUTES",
                self.DEFAULT_WEB_AUTH_REFRESH_TOKEN_EXPIRE_MINUTES,
                self.ENV_PREFIX,
            )
        )

    @WEB_AUTH_REFRESH_TOKEN_EXPIRE_MINUTES.setter
    def WEB_AUTH_REFRESH_TOKEN_EXPIRE_MINUTES(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_WEB_REFRESH_TOKEN_EXPIRE_MINUTES"] = str(value)

    @property
    def WEB_TITLE(self) -> str:
        return get_env("WEB_TITLE", self.DEFAULT_WEB_TITLE, self.ENV_PREFIX)

    @WEB_TITLE.setter
    def WEB_TITLE(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_WEB_TITLE"] = value

    @property
    def WEB_JARGON(self) -> str:
        return get_env("WEB_JARGON", self.DEFAULT_WEB_JARGON, self.ENV_PREFIX)

    @WEB_JARGON.setter
    def WEB_JARGON(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_WEB_JARGON"] = value

    @property
    def WEB_HOMEPAGE_INTRO(self) -> str:
        return get_env(
            "WEB_HOMEPAGE_INTRO", self.DEFAULT_WEB_HOMEPAGE_INTRO, self.ENV_PREFIX
        )

    @WEB_HOMEPAGE_INTRO.setter
    def WEB_HOMEPAGE_INTRO(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_WEB_HOMEPAGE_INTRO"] = value

    @property
    def LLM_ASSISTANT_NAME(self) -> str:
        default = self.DEFAULT_LLM_ASSISTANT_NAME
        if default == "":
            default = self.ROOT_GROUP_NAME
        return get_env("LLM_ASSISTANT_NAME", default, self.ENV_PREFIX)

    @LLM_ASSISTANT_NAME.setter
    def LLM_ASSISTANT_NAME(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_LLM_ASSISTANT_NAME"] = value

    @property
    def LLM_ASSISTANT_ASCII_ART(self) -> str:
        return get_env(
            "LLM_ASSISTANT_ASCII_ART",
            self.DEFAULT_LLM_ASSISTANT_ASCII_ART,
            self.ENV_PREFIX,
        )

    @LLM_ASSISTANT_ASCII_ART.setter
    def LLM_ASSISTANT_ASCII_ART(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_LLM_ASSISTANT_ASCII_ART"] = value

    @property
    def LLM_ASSISTANT_JARGON(self) -> str:
        default = self.DEFAULT_LLM_ASSISTANT_JARGON
        if default == "":
            default = self.ROOT_GROUP_DESCRIPTION
        return get_env(
            "LLM_ASSISTANT_JARGON",
            default,
            self.ENV_PREFIX,
        )

    @LLM_ASSISTANT_JARGON.setter
    def LLM_ASSISTANT_JARGON(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_LLM_ASSISTANT_JARGON"] = value

    @property
    def LLM_HISTORY_DIR(self) -> str:
        default = self.DEFAULT_LLM_HISTORY_DIR
        if default == "":
            default = os.path.expanduser(
                os.path.join("~", f".{self.ROOT_GROUP_NAME}", "llm-history")
            )
        return get_env(
            "LLM_HISTORY_DIR",
            default,
            self.ENV_PREFIX,
        )

    @LLM_HISTORY_DIR.setter
    def LLM_HISTORY_DIR(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_LLM_HISTORY_DIR"] = value

    @property
    def LLM_NOTE_FILE(self) -> str:
        default = self.DEFAULT_LLM_NOTE_FILE
        if default == "":
            default = os.path.expanduser(
                os.path.join("~", f".{self.ROOT_GROUP_NAME}", "notes.json")
            )
        return get_env(
            "LLM_NOTE_FILE",
            default,
            self.ENV_PREFIX,
        )

    @LLM_NOTE_FILE.setter
    def LLM_NOTE_FILE(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_LLM_NOTE_FILE"] = value

    @property
    def LLM_MODEL(self) -> str | None:
        value = get_env("LLM_MODEL", self.DEFAULT_LLM_MODEL, self.ENV_PREFIX)
        return None if value == "" else value

    @LLM_MODEL.setter
    def LLM_MODEL(self, value: str | None):
        if value is None:
            if f"{self.ENV_PREFIX}_LLM_MODEL" in os.environ:
                del os.environ[f"{self.ENV_PREFIX}_LLM_MODEL"]
        else:
            os.environ[f"{self.ENV_PREFIX}_LLM_MODEL"] = value

    @property
    def LLM_BASE_URL(self) -> str | None:
        value = get_env("LLM_BASE_URL", self.DEFAULT_LLM_BASE_URL, self.ENV_PREFIX)
        return None if value == "" else value

    @LLM_BASE_URL.setter
    def LLM_BASE_URL(self, value: str | None):
        if value is None:
            if f"{self.ENV_PREFIX}_LLM_BASE_URL" in os.environ:
                del os.environ[f"{self.ENV_PREFIX}_LLM_BASE_URL"]
        else:
            os.environ[f"{self.ENV_PREFIX}_LLM_BASE_URL"] = value

    @property
    def LLM_API_KEY(self) -> str | None:
        value = get_env("LLM_API_KEY", self.DEFAULT_LLM_API_KEY, self.ENV_PREFIX)
        return None if value == "" else value

    @LLM_API_KEY.setter
    def LLM_API_KEY(self, value: str | None):
        if value is None:
            if f"{self.ENV_PREFIX}_LLM_API_KEY" in os.environ:
                del os.environ[f"{self.ENV_PREFIX}_LLM_API_KEY"]
        else:
            os.environ[f"{self.ENV_PREFIX}_LLM_API_KEY"] = value

    @property
    def LLM_MAX_REQUESTS_PER_MINUTE(self) -> int:
        """
        Maximum number of LLM requests allowed per minute.
        Default is conservative to accommodate free-tier LLM providers.
        """
        return int(
            get_env(
                ["LLM_MAX_REQUEST_PER_MINUTE", "LLM_MAX_REQUESTS_PER_MINUTE"],
                self.DEFAULT_LLM_MAX_REQUESTS_PER_MINUTE,
                self.ENV_PREFIX,
            )
        )

    @LLM_MAX_REQUESTS_PER_MINUTE.setter
    def LLM_MAX_REQUESTS_PER_MINUTE(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_LLM_MAX_REQUESTS_PER_MINUTE"] = str(value)

    @property
    def LLM_MAX_TOKENS_PER_MINUTE(self) -> int:
        """
        Maximum number of LLM tokens allowed per minute.
        Default is conservative to accommodate free-tier LLM providers.
        """
        return int(
            get_env(
                ["LLM_MAX_TOKEN_PER_MINUTE", "LLM_MAX_TOKENS_PER_MINUTE"],
                self.DEFAULT_LLM_MAX_TOKENS_PER_MINUTE,
                self.ENV_PREFIX,
            )
        )

    @LLM_MAX_TOKENS_PER_MINUTE.setter
    def LLM_MAX_TOKENS_PER_MINUTE(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_LLM_MAX_TOKENS_PER_MINUTE"] = str(value)

    @property
    def LLM_MAX_TOKENS_PER_REQUEST(self) -> int:
        """Maximum number of tokens allowed per individual LLM request."""
        return int(
            get_env(
                ["LLM_MAX_TOKEN_PER_REQUEST", "LLM_MAX_TOKENS_PER_REQUEST"],
                self.DEFAULT_LLM_MAX_TOKENS_PER_REQUEST,
                self.ENV_PREFIX,
            )
        )

    @LLM_MAX_TOKENS_PER_REQUEST.setter
    def LLM_MAX_TOKENS_PER_REQUEST(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_LLM_MAX_TOKENS_PER_REQUEST"] = str(value)

    @property
    def LLM_THROTTLE_SLEEP(self) -> float:
        """Number of seconds to sleep when throttling is required."""
        return float(
            get_env(
                "LLM_THROTTLE_SLEEP",
                self.DEFAULT_LLM_THROTTLE_SLEEP,
                self.ENV_PREFIX,
            )
        )

    @LLM_THROTTLE_SLEEP.setter
    def LLM_THROTTLE_SLEEP(self, value: float):
        os.environ[f"{self.ENV_PREFIX}_LLM_THROTTLE_SLEEP"] = str(value)

    @property
    def LLM_HISTORY_SUMMARIZATION_WINDOW(self) -> int:
        return int(
            get_env(
                "LLM_HISTORY_SUMMARIZATION_WINDOW",
                self.DEFAULT_LLM_HISTORY_SUMMARIZATION_WINDOW,
                self.ENV_PREFIX,
            )
        )

    @LLM_HISTORY_SUMMARIZATION_WINDOW.setter
    def LLM_HISTORY_SUMMARIZATION_WINDOW(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_LLM_HISTORY_SUMMARIZATION_WINDOW"] = str(value)

    @property
    def LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD(self) -> int:
        default = self.DEFAULT_LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD
        if default == "":
            default = str(self._get_max_threshold(0.6))

        threshold = int(
            get_env(
                "LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD",
                default,
                self.ENV_PREFIX,
            )
        )
        return limit_token_threshold(
            threshold,
            0.6,
            self.LLM_MAX_TOKENS_PER_MINUTE,
            self.LLM_MAX_TOKENS_PER_REQUEST,
        )

    @LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD.setter
    def LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD"] = (
            str(value)
        )

    @property
    def LLM_REPO_ANALYSIS_EXTRACTION_TOKEN_THRESHOLD(self) -> int:
        default = self.DEFAULT_LLM_REPO_ANALYSIS_EXTRACTION_TOKEN_THRESHOLD
        if default == "":
            default = str(self._get_max_threshold(0.4))

        threshold = int(
            get_env(
                "LLM_REPO_ANALYSIS_EXTRACTION_TOKEN_THRESHOLD",
                default,
                self.ENV_PREFIX,
            )
        )
        return limit_token_threshold(
            threshold,
            0.4,
            self.LLM_MAX_TOKENS_PER_MINUTE,
            self.LLM_MAX_TOKENS_PER_REQUEST,
        )

    @LLM_REPO_ANALYSIS_EXTRACTION_TOKEN_THRESHOLD.setter
    def LLM_REPO_ANALYSIS_EXTRACTION_TOKEN_THRESHOLD(self, value: int):
        os.environ[
            f"{self.ENV_PREFIX}_LLM_REPO_ANALYSIS_EXTRACTION_TOKEN_THRESHOLD"
        ] = str(value)

    @property
    def LLM_REPO_ANALYSIS_SUMMARIZATION_TOKEN_THRESHOLD(self) -> int:
        default = self.DEFAULT_LLM_REPO_ANALYSIS_SUMMARIZATION_TOKEN_THRESHOLD
        if default == "":
            default = str(self._get_max_threshold(0.4))

        threshold = int(
            get_env(
                "LLM_REPO_ANALYSIS_SUMMARIZATION_TOKEN_THRESHOLD",
                default,
                self.ENV_PREFIX,
            )
        )
        return limit_token_threshold(
            threshold,
            0.4,
            self.LLM_MAX_TOKENS_PER_MINUTE,
            self.LLM_MAX_TOKENS_PER_REQUEST,
        )

    @LLM_REPO_ANALYSIS_SUMMARIZATION_TOKEN_THRESHOLD.setter
    def LLM_REPO_ANALYSIS_SUMMARIZATION_TOKEN_THRESHOLD(self, value: int):
        os.environ[
            f"{self.ENV_PREFIX}_LLM_REPO_ANALYSIS_SUMMARIZATION_TOKEN_THRESHOLD"
        ] = str(value)

    @property
    def LLM_FILE_ANALYSIS_TOKEN_THRESHOLD(self) -> int:
        default = self.DEFAULT_LLM_FILE_ANALYSIS_TOKEN_THRESHOLD
        if default == "":
            default = str(self._get_max_threshold(0.4))

        threshold = int(
            get_env(
                "LLM_FILE_ANALYSIS_TOKEN_THRESHOLD",
                default,
                self.ENV_PREFIX,
            )
        )
        return limit_token_threshold(
            threshold,
            0.4,
            self.LLM_MAX_TOKENS_PER_MINUTE,
            self.LLM_MAX_TOKENS_PER_REQUEST,
        )

    @LLM_FILE_ANALYSIS_TOKEN_THRESHOLD.setter
    def LLM_FILE_ANALYSIS_TOKEN_THRESHOLD(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_LLM_FILE_ANALYSIS_TOKEN_THRESHOLD"] = str(value)

    @property
    def LLM_PROMPT_DIR(self) -> str:
        default = self.DEFAULT_LLM_PROMPT_DIR
        if default == "":
            default = os.path.join(f".{self.ROOT_GROUP_NAME}", "llm", "prompt")
        return get_env(
            "LLM_PROMPT_DIR",
            default,
            self.ENV_PREFIX,
        )

    @LLM_PROMPT_DIR.setter
    def LLM_PROMPT_DIR(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_LLM_PROMPT_DIR"] = value

    def _get_max_threshold(self, factor: float) -> int:
        return get_max_token_threshold(
            factor, self.LLM_MAX_TOKENS_PER_MINUTE, self.LLM_MAX_TOKENS_PER_REQUEST
        )

    @property
    def RAG_EMBEDDING_API_KEY(self) -> str | None:
        value = get_env(
            "RAG_EMBEDDING_API_KEY", self.DEFAULT_RAG_EMBEDDING_API_KEY, self.ENV_PREFIX
        )
        return None if value == "" else value

    @RAG_EMBEDDING_API_KEY.setter
    def RAG_EMBEDDING_API_KEY(self, value: str | None):
        if value is None:
            if f"{self.ENV_PREFIX}_RAG_EMBEDDING_API_KEY" in os.environ:
                del os.environ[f"{self.ENV_PREFIX}_RAG_EMBEDDING_API_KEY"]
        else:
            os.environ[f"{self.ENV_PREFIX}_RAG_EMBEDDING_API_KEY"] = value

    @property
    def RAG_EMBEDDING_BASE_URL(self) -> str | None:
        value = get_env(
            "RAG_EMBEDDING_BASE_URL",
            self.DEFAULT_RAG_EMBEDDING_BASE_URL,
            self.ENV_PREFIX,
        )
        return None if value == "" else value

    @RAG_EMBEDDING_BASE_URL.setter
    def RAG_EMBEDDING_BASE_URL(self, value: str | None):
        if value is None:
            if f"{self.ENV_PREFIX}_RAG_EMBEDDING_BASE_URL" in os.environ:
                del os.environ[f"{self.ENV_PREFIX}_RAG_EMBEDDING_BASE_URL"]
        else:
            os.environ[f"{self.ENV_PREFIX}_RAG_EMBEDDING_BASE_URL"] = value

    @property
    def RAG_EMBEDDING_MODEL(self) -> str:
        return get_env(
            "RAG_EMBEDDING_MODEL", self.DEFAULT_RAG_EMBEDDING_MODEL, self.ENV_PREFIX
        )

    @RAG_EMBEDDING_MODEL.setter
    def RAG_EMBEDDING_MODEL(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_RAG_EMBEDDING_MODEL"] = value

    @property
    def RAG_CHUNK_SIZE(self) -> int:
        return int(
            get_env("RAG_CHUNK_SIZE", self.DEFAULT_RAG_CHUNK_SIZE, self.ENV_PREFIX)
        )

    @RAG_CHUNK_SIZE.setter
    def RAG_CHUNK_SIZE(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_RAG_CHUNK_SIZE"] = str(value)

    @property
    def RAG_OVERLAP(self) -> int:
        return int(get_env("RAG_OVERLAP", self.DEFAULT_RAG_OVERLAP, self.ENV_PREFIX))

    @RAG_OVERLAP.setter
    def RAG_OVERLAP(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_RAG_OVERLAP"] = str(value)

    @property
    def RAG_MAX_RESULT_COUNT(self) -> int:
        return int(
            get_env(
                "RAG_MAX_RESULT_COUNT",
                self.DEFAULT_RAG_MAX_RESULT_COUNT,
                self.ENV_PREFIX,
            )
        )

    @RAG_MAX_RESULT_COUNT.setter
    def RAG_MAX_RESULT_COUNT(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_RAG_MAX_RESULT_COUNT"] = str(value)

    @property
    def SEARCH_INTERNET_METHOD(self) -> str:
        """Either serpapi or searxng"""
        return get_env(
            "SEARCH_INTERNET_METHOD",
            self.DEFAULT_SEARCH_INTERNET_METHOD,
            self.ENV_PREFIX,
        )

    @SEARCH_INTERNET_METHOD.setter
    def SEARCH_INTERNET_METHOD(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_SEARCH_INTERNET_METHOD"] = value

    @property
    def BRAVE_API_KEY(self) -> str:
        return os.getenv("BRAVE_API_KEY", self.DEFAULT_BRAVE_API_KEY)

    @BRAVE_API_KEY.setter
    def BRAVE_API_KEY(self, value: str):
        os.environ["BRAVE_API_KEY"] = value

    @property
    def BRAVE_API_SAFE(self) -> str:
        return get_env("BRAVE_API_SAFE", self.DEFAULT_BRAVE_API_SAFE, self.ENV_PREFIX)

    @BRAVE_API_SAFE.setter
    def BRAVE_API_SAFE(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_BRAVE_API_SAFE"] = value

    @property
    def BRAVE_API_LANG(self) -> str:
        return get_env("BRAVE_API_LANG", self.DEFAULT_BRAVE_API_LANG, self.ENV_PREFIX)

    @BRAVE_API_LANG.setter
    def BRAVE_API_LANG(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_BRAVE_API_LANG"] = value

    @property
    def SERPAPI_KEY(self) -> str:
        return os.getenv("SERPAPI_KEY", self.DEFAULT_SERPAPI_KEY)

    @SERPAPI_KEY.setter
    def SERPAPI_KEY(self, value: str):
        os.environ["SERPAPI_KEY"] = value

    @property
    def SERPAPI_SAFE(self) -> str:
        return get_env("SERPAPI_SAFE", self.DEFAULT_SERPAPI_SAFE, self.ENV_PREFIX)

    @SERPAPI_SAFE.setter
    def SERPAPI_SAFE(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_SERPAPI_SAFE"] = value

    @property
    def SERPAPI_LANG(self) -> str:
        return get_env("SERPAPI_LANG", self.DEFAULT_SERPAPI_LANG, self.ENV_PREFIX)

    @SERPAPI_LANG.setter
    def SERPAPI_LANG(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_SERPAPI_LANG"] = value

    @property
    def SEARXNG_PORT(self) -> int:
        return int(get_env("SEARXNG_PORT", self.DEFAULT_SEARXNG_PORT, self.ENV_PREFIX))

    @SEARXNG_PORT.setter
    def SEARXNG_PORT(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_SEARXNG_PORT"] = str(value)

    @property
    def SEARXNG_BASE_URL(self) -> str:
        default = self.DEFAULT_SEARXNG_BASE_URL
        if default == "":
            default = f"http://localhost:{self.SEARXNG_PORT}"
        return get_env("SEARXNG_BASE_URL", default, self.ENV_PREFIX)

    @SEARXNG_BASE_URL.setter
    def SEARXNG_BASE_URL(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_SEARXNG_BASE_URL"] = value

    @property
    def SEARXNG_SAFE(self) -> int:
        return int(get_env("SEARXNG_SAFE", self.DEFAULT_SEARXNG_SAFE, self.ENV_PREFIX))

    @SEARXNG_SAFE.setter
    def SEARXNG_SAFE(self, value: int):
        os.environ[f"{self.ENV_PREFIX}_SEARXNG_SAFE"] = str(value)

    @property
    def SEARXNG_LANG(self) -> str:
        return get_env("SEARXNG_LANG", self.DEFAULT_SEARXNG_LANG, self.ENV_PREFIX)

    @SEARXNG_LANG.setter
    def SEARXNG_LANG(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_SEARXNG_LANG"] = value

    @property
    def BANNER(self) -> str:
        return fstring_format(
            get_env("BANNER", self.DEFAULT_BANNER, self.ENV_PREFIX),
            {"VERSION": self.VERSION},
        )

    @BANNER.setter
    def BANNER(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_BANNER"] = value

    @property
    def LLM_SHOW_TOOL_CALL_DETAIL(self) -> bool:
        return to_boolean(
            get_env(
                "LLM_SHOW_TOOL_CALL_DETAIL",
                self.DEFAULT_LLM_SHOW_TOOL_CALL_DETAIL,
                self.ENV_PREFIX,
            )
        )

    @LLM_SHOW_TOOL_CALL_DETAIL.setter
    def LLM_SHOW_TOOL_CALL_DETAIL(self, value: bool):
        os.environ[f"{self.ENV_PREFIX}_LLM_SHOW_TOOL_CALL_DETAIL"] = (
            "1" if value else "0"
        )

    @property
    def LLM_SHOW_TOOL_CALL_RESULT(self) -> bool:
        return to_boolean(
            get_env(
                "LLM_SHOW_TOOL_CALL_RESULT",
                self.DEFAULT_LLM_SHOW_TOOL_CALL_RESULT,
                self.ENV_PREFIX,
            )
        )

    @LLM_SHOW_TOOL_CALL_RESULT.setter
    def LLM_SHOW_TOOL_CALL_RESULT(self, value: bool):
        os.environ[f"{self.ENV_PREFIX}_LLM_SHOW_TOOL_CALL_RESULT"] = (
            "1" if value else "0"
        )

    @property
    def USE_TIKTOKEN(self) -> bool:
        return to_boolean(
            get_env("USE_TIKTOKEN", self.DEFAULT_USE_TIKTOKEN, self.ENV_PREFIX)
        )

    @USE_TIKTOKEN.setter
    def USE_TIKTOKEN(self, value: bool):
        os.environ[f"{self.ENV_PREFIX}_USE_TIKTOKEN"] = "1" if value else "0"

    @property
    def TIKTOKEN_ENCODING_NAME(self) -> str:
        return get_env(
            ["TIKTOKEN_ENCODING", "TIKTOKEN_ENCODING_NAME"],
            self.DEFAULT_TIKTOKEN_ENCODING_NAME,
            self.ENV_PREFIX,
        )

    @TIKTOKEN_ENCODING_NAME.setter
    def TIKTOKEN_ENCODING_NAME(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_TIKTOKEN_ENCODING_NAME"] = value

    @property
    def MCP_CONFIG_FILE(self) -> str:
        return get_env("MCP_CONFIG_FILE", self.DEFAULT_MCP_CONFIG_FILE, self.ENV_PREFIX)

    @MCP_CONFIG_FILE.setter
    def MCP_CONFIG_FILE(self, value: str):
        os.environ[f"{self.ENV_PREFIX}_MCP_CONFIG_FILE"] = value


CFG = Config()



from fastapi.testclient import TestClient
from my_app_name.main import app
from my_app_name.module.auth.service.user.user_service_factory import user_service
from my_app_name.module.gateway.util.view import render_content


async def test_homepage():
    current_user = await user_service.get_current_user(access_token=None)
    client = TestClient(app, base_url="http://localhost")
    response = client.get("/")
    assert response.status_code == 200
    assert response.text == render_content(
        "homepage.html",
        page_name="gateway.home",
        current_user=current_user,
    ).body.decode("utf-8")



from typing import Any

import requests

from zrb.config.config import CFG


def search_internet(
    query: str,
    page: int = 1,
    safe_search: int | None = None,
    language: str | None = None,
) -> dict[str, Any]:
    """
    Performs a live internet search using SearXNG, an aggregator that combines results from multiple search engines.

    **WHEN TO USE:**
    - To gather diverse perspectives or information from across the web.
    - To retrieve the latest data, documentation, or public resources.

    **ARGS:**
    - `query`: The search string or question.
    - `page`: Result page number (default 1).
    """
    if safe_search is None:
        safe_search = CFG.SEARXNG_SAFE
    if language is None:
        language = CFG.SEARXNG_LANG

    user_agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"

    response = requests.get(
        url=f"{CFG.SEARXNG_BASE_URL}/search",
        headers={"User-Agent": user_agent},
        params={
            "q": query,
            "format": "json",
            "pageno": page,
            "safesearch": safe_search,
            "language": language,
        },
    )
    if response.status_code != 200:
        raise Exception(
            f"Error: Unable to retrieve search results (status code: {response.status_code})"
        )
    return response.json()



import string

from prompt_toolkit.application import get_app
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.widgets import TextArea


def create_output_keybindings(input_field: TextArea) -> KeyBindings:
    kb = KeyBindings()

    @kb.add("escape")
    def _(event):
        get_app().layout.focus(input_field)

    @kb.add("c-c")
    def _(event):
        # Copy selection to clipboard
        if event.current_buffer.selection_state:
            data = event.current_buffer.copy_selection()
            event.app.clipboard.set_data(data)
        get_app().layout.focus(input_field)

    def redirect_focus(event):
        get_app().layout.focus(input_field)
        input_field.buffer.insert_text(event.data)

    for char in string.printable:
        # Skip control characters (Tab, Newline, etc.)
        #  to preserve navigation/standard behavior
        if char in "\t\n\r\x0b\x0c":
            continue
        kb.add(char)(redirect_focus)

    return kb



from typing import Any

from zrb.llm.agent.agent import create_agent, run_agent
from zrb.llm.config.config import LLMConfig
from zrb.llm.config.config import llm_config as default_config
from zrb.llm.config.limiter import LLMLimiter
from zrb.llm.config.limiter import llm_limiter as default_limiter
from zrb.llm.history_manager.any_history_manager import AnyHistoryManager
from zrb.llm.history_manager.file_history_manager import FileHistoryManager


def create_sub_agent_tool(
    name: str,
    description: str,
    model: str | None = None,
    system_prompt: str = "",
    tools: list = [],
    llm_config: LLMConfig | None = None,
    llm_limitter: LLMLimiter | None = None,
    history_manager: AnyHistoryManager | None = None,
    conversation_name: str = "sub_agent_default",
) -> Any:
    """
    Creates a Tool that invokes a sub-agent.
    The sub-agent manages its own persistent history via HistoryManager
    and handles approval via CLI fallback if needed.
    """
    config = llm_config or default_config
    limiter = llm_limitter or default_limiter
    manager = history_manager or FileHistoryManager(history_dir="~/.llm_chat/subagents")
    final_model = model or config.model

    agent_instance = None

    async def run_sub_agent(prompt: str) -> str:
        nonlocal agent_instance
        if agent_instance is None:
            agent_instance = create_agent(
                model=final_model,
                system_prompt=system_prompt,
                tools=tools,
                yolo=False,
            )

        # Load persistent history
        history = manager.load(conversation_name)

        # Execute agent with blocking confirmation loop for approvals
        result, new_history = await run_agent(
            agent=agent_instance,
            message=prompt,
            message_history=history,
            limiter=limiter,
        )

        # Save updated history
        manager.update(conversation_name, new_history)
        manager.save(conversation_name)

        return str(result)

    run_sub_agent.__name__ = name
    run_sub_agent.__doc__ = (
        f"DELEGATION TOOL: {description}\n\n"
        "Use this tool to delegate complex, multi-step sub-tasks to a specialized agent. "
        "The sub-agent has its own memory and can perform its own tool calls."
        "\n\n**ARGS:**"
        "\n- `prompt`: The clear and detailed objective or instruction for the sub-agent."
    )
    return run_sub_agent



# MyEntity related methods


@abstractmethod
async def get_my_entity_by_id(self, my_entity_id: str) -> MyEntityResponse:
    """Get my entity by id"""


@abstractmethod
async def get_my_entities(
    self,
    page: int = 1,
    page_size: int = 10,
    sort: str | None = None,
    filter: str | None = None,
) -> MultipleMyEntityResponse:
    """Get my entities by filter and sort"""


@abstractmethod
async def create_my_entity_bulk(
    self, data: list[MyEntityCreateWithAudit]
) -> list[MyEntityResponse]:
    """Create new my entities"""


@abstractmethod
async def create_my_entity(self, data: MyEntityCreateWithAudit) -> MyEntityResponse:
    """Create a new my entities"""


@abstractmethod
async def update_my_entity_bulk(
    self, my_entity_ids: list[str], data: MyEntityUpdateWithAudit
) -> MyEntityResponse:
    """Update some my entities"""


@abstractmethod
async def update_my_entity(
    self, my_entity_id: str, data: MyEntityUpdateWithAudit
) -> MyEntityResponse:
    """Update a my entity"""


@abstractmethod
async def delete_my_entity_bulk(
    self, my_entity_ids: str, deleted_by: str
) -> MyEntityResponse:
    """Delete some my entities"""


@abstractmethod
async def delete_my_entity(
    self, my_entity_id: str, deleted_by: str
) -> MyEntityResponse:
    """Delete a my entity"""
